from simobjects.box import *
import numpy as np

mesh_box = jb.MeshMaterial(texture_map='http://127.0.0.1:8887/metal.png', roughness=1, metalness=1, opacity=0.5)
box = jb.Box(name="movingBox", width=0.3, height=0.4, depth=0.2, mesh_material=mesh_box)

light1 = jb.PointLight(name="light1", color="white", intensity=6, position=[2, 0, 3])
light2 = jb.PointLight(name="light2", color="white", intensity=6, position=[-2, 0, 3])
light3 = jb.PointLight(name="light3", color="white", intensity=6, position=[0, 0, 3])
light4 = jb.PointLight(name="light4", color="white", intensity=6, position=[0, 1, 1])

lin_velocity = jb.Vector(name="linVelocity", color="cyan")
ang_velocity = jb.Vector(name="angVelocity", color="orange")

sim = jb.Simulation([box, light1, light2, light3, light4, lin_velocity, ang_velocity])

dt = 0.01

pose = lambda t: jb.Utils.trn([np.cos(0.3 * t), np.sin(0.5 * t), 0.5 * np.sin(0.5 * t) + 1]) @ jb.Utils.rotz(
    0.2 * t) @ jb.Utils.rotx(0.1 * t)

for i in range(3000):
    htm = pose(i * dt)
    dot_htm = (pose((i + 1) * dt) - pose((i - 1) * dt)) / (2 * dt)
    mat_s = dot_htm[0:3, 0:3] @ np.transpose(htm[0:3, 0:3])

    lin_v = dot_htm[0:3, 3]
    lin_w = [2 * mat_s[2][1], 2 * mat_s[0][2], 2 * mat_s[1][0]]

    box.add_ani_frame(time=i * dt, htm=htm)
    lin_velocity.add_ani_frame(time=i * dt, origin=htm[0:3, 3], vector=lin_v)
    ang_velocity.add_ani_frame(time=i * dt, origin=htm[0:3, 3], vector=lin_w)

sim.run()



__version__ = '0.0.21'

from .robot import *
from .demo import *
from .utils import *
from uaibot.robot.links import *
from .simulation import *
from sfm import *
from uaibot.simobjects.ball import *
from uaibot.simobjects.box import *
from uaibot.simobjects.smoothbox import *
from uaibot.simobjects.cylinder import *
from uaibot.simobjects.cone import *
from uaibot.graphics.meshmaterial import *
from uaibot.simobjects.pointlight import *
from uaibot.graphics.texture import *
from uaibot.simobjects.frame import *
from uaibot.graphics.model3d import *
from uaibot.simobjects.pointcloud import *
from uaibot.simobjects.vector import *
from uaibot.simobjects.rigidobject import *
from uaibot.simobjects.group import *
from uaibot.simobjects.htmldiv import *
from uaibot.algorithm.rrt import *


from utils import *
import numpy as np
from graphics.meshmaterial import *


class Ball:
    """
  A ball object.

  Parameters
  ----------
  htm : 4x4 numpy array or 4x4 nested list
      The object's configuration.
      (default: the same as the current HTM).

  name : string
      The object's name.
      (default: 'genBall').

  radius : positive float
      The object's radius, in meters.
      (default: 1).    

  mass : positive float
      The object's mass, in kg.
      (default: 1).  

  color : string
      The object's color, a HTML - compatible string.
      (default: "red").

  opacity : float between 0 and 1
      The opacity. 1 = fully opaque, and 0 = transparent.

  mesh_material: 'MeshMaterial' object
      The object mesh material. If set to 'None', the default is used.
      (default: None).
  """

    #######################################
    # Attributes
    #######################################

    @property
    def radius(self):
        """The ball radius, in meters."""
        return self._radius

    @property
    def name(self):
        """Name of the object."""
        return self._name

    @property
    def htm(self):
        """Object pose. A 4x4 homogeneous transformation matrix written is scenario coordinates."""
        return np.matrix(self._htm)

    @property
    def mass(self):
        """Mass of the object, in kg."""
        return self._mass

    @property
    def color(self):
        """Color of the object"""
        return self.mesh_material.color

    @property
    def mesh_material(self):
        """Mesh properties of the object."""
        return self._mesh_material

    @property
    def volume(self):
        """The volume of the object, in m³."""
        return self._volume

    #######################################
    # Constructor
    #######################################

    def __init__(self, htm=np.identity(4), name="", radius=1, mass=1, color="red", opacity=1,
                 mesh_material=None):

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_number(mass) or mass < 0:
            raise Exception("The parameter 'mass' should be a positive float.")

        if not Utils.is_a_number(radius) or radius < 0:
            raise Exception("The parameter 'radius' should be a positive float.")

        if name=="":
            name="var_ball_id_"+str(id(self))

        if not (Utils.is_a_name(name)):
            raise Exception(
                "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

        if not Utils.is_a_color(color):
            raise Exception("The parameter 'color' should be a HTML-compatible color.")

        if not ((mesh_material is None) or (Utils.get_uaibot_type(mesh_material) == "uaibot.MeshMaterial")):
            raise Exception(
                "The parameter 'mesh_material' should be either 'None' or a 'uaibot.MeshMaterial' object.")

        if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
            raise Exception("The parameter 'opacity' should be a float between 0 and 1.")
        # end error handling

        self._radius = radius
        self._htm = np.matrix(htm)
        self._name = name
        self._mass = 1
        self._frames = []
        self._volume = (4 / 3) * np.pi * self.radius * self.radius * self.radius
        self._max_time = 0

        if mesh_material is None:
            self._mesh_material = MeshMaterial(color=color, opacity=opacity)
        else:
            self._mesh_material = mesh_material

        # Set initial total configuration
        self.set_ani_frame(self._htm)

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Ball with name '" + self.name + "': \n\n"
        string += " Radius (m): " + str(self.radius) + "\n"
        string += " Color: " + str(self.color) + "\n"
        string += " Mass (kg): " + str(self.mass) + "\n"
        string += " HTM: \n" + str(self.htm) + "\n"

        return string

    #######################################
    # Methods
    #######################################

    def add_ani_frame(self, time, htm=None):
        """
    Add a single configuration to the object's animation queue.

    Parameters
    ----------
    time: positive float
        The timestamp of the animation frame, in seconds.
    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """
        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_number(time) or time < 0:
            raise Exception("The parameter 'time' should be a positive float.")
        # end error handling

        f = [time, np.around(htm[0,0],4), np.around(htm[0,1],4), np.around(htm[0,2],4), np.around(htm[0,3],4),
             np.around(htm[1,0],4), np.around(htm[1,1],4), np.around(htm[1,2],4), np.around(htm[1,3],4),
             np.around(htm[2,0],4), np.around(htm[2,1],4), np.around(htm[2,2],4), np.around(htm[2,3],4),
             0, 0, 0, 1]

        self._htm = htm
        self._frames.append(f)
        self._max_time = max(self._max_time, time)

    # Set config. Restart animation queue
    def set_ani_frame(self, htm=None):
        """
    Reset object's animation queue and add a single configuration to the 
    object's animation queue.

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """

        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        # end error handling

        self._frames = []
        self.add_ani_frame(0, htm)
        self._max_time = 0

    def gen_code(self):
        """Generate code for injection."""

        string = "\n"
        string += "//BEGIN DECLARATION OF THE BALL '" + self.name + "'\n\n"
        string += self.mesh_material.gen_code(self.name) + "\n"
        string += "const var_" + self.name + " = new Ball(" + str(self.radius) + "," + str(
            self._frames) + ", material_" + self.name + ");\n"
        string += "sceneElements.push(var_" + self.name + ");\n"
        string += "//USER INPUT GOES HERE"

        return string

    # Compute inertia matrix with respect to the inertia frame
    def inertia_matrix(self, htm=None):
        """
    The 3D inertia matrix of the object, written in the world frame.
    Assume that the transformation between the word frame and the object frame is 'htm'.

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration for which the inertia matrix will be computed
        (default: the same as the current HTM).

    Returns
    -------
     inertia_matrix : 3x3 numpy array
        The 3D inertia matrix.
    """

        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The optional parameter 'htm' should be a 4x4 homogeneous transformation matrix.")
        # end error handling

        S = Utils.S(htm[0:3, 3])
        I = (2 / 5) * self.mass * (self._radius * self._radius)

        return I * np.identity(3) - self.mass * S * S

    def copy(self):
        """Return a deep copy of the object, without copying the animation frames."""
        return Ball(self.htm, self.name + "_copy", self.radius, self.mass, self.color)

    def aabb(self):
        """
    Compute the width, depth and height of an axis aligned bounding box (aabb) that
    covers the object. It also considers the current orientation.

    Returns
    -------
     width : positive float
        The width of the box, in meters.

     depth : positive float
        The depth of the box, in meters.

     height : positive float
        The depth of the box, in meters.
    """
        return 2 * self.radius, 2 * self.radius, 2 * self.radius

    #Generate samples
    def generate_samples(self, delta=0.025):

        N = round(self.radius / delta)+1
        P = np.matrix(np.zeros((3, 0)))
        for i in range(N):
            phi =  np.pi*(i/N)
            M = round(N*np.sin(phi))+1
            for j in range(M):
                theta = 2 * np.pi *(j/M)
                x = self.radius * np.sin(phi) * np.cos(theta)
                y = self.radius * np.sin(phi) * np.sin(theta)
                z = self.radius * np.cos(phi)
                P = np.block([P, np.matrix([x, y,  z]).transpose()])



        for i in range(np.shape(P)[1]):
            P[:,i] = self.htm[0:3,0:3]*P[:,i]+self.htm[0:3,-1]

        return P



    # Compute the projection of a point into an object
    def projection(self, point, htm=None):
        """
    The projection of a point in the object, that is, the
    closest point in the object to a point 'point'.

    Parameters
    ----------
    point : 3D vector
        The point for which the projection will be computed.

    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration
        (default: the same as the current HTM).            

    Returns
    -------
     proj_point : 3D vector
        The projection of the point 'point' in the object.

     d : positive float
        The distance between the object and 'point'.
    """

        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The optional parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_vector(point, 3):
            raise Exception("The parameter 'point' should be a 3D vector.")

        # end error handling

        dd = np.linalg.norm(point - htm[0:3, 3])
        d = max(dd-self.radius,0)

        if d == 0:
            return np.matrix(point).reshape((3,1)), d
        else:
            cp = htm[0:3, 3] + self.radius * (np.matrix(point).reshape((3,1)) - htm[0:3, 3]) / dd
            return cp, d





from simulation import *
import numpy as np
from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *
from simobjects.pointlight import *
from simobjects.pointcloud import *
from simobjects.frame import *
from utils import *
import robot as rb
from cvxopt import matrix, solvers
from graphics.meshmaterial import *
import sys


def _control_demo_3_old():
    solvers.options['show_progress'] = False

    # Create obstacle object and robots

    # Create simulation and add objects to the scene
    robot = rb.Robot.create_kuka_kr5(np.identity(4), "robot")
    mesh_obstacle = MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="green")
    obstacle = Box(htm=Utils.trn([0.09, 0.5, 0.6]), name='obstacle', width=0.14, height=0.14, depth=0.14,
                   mesh_material=mesh_obstacle)


    desired_eef_frame = Frame(axis_color=["darkred", "green", "darkblue"], size=0.33)

    sim = Simulation.create_sim_factory([robot, obstacle, desired_eef_frame])

    # Starting pose for the end-effector
    htm0 = Utils.trn([-0.2, 0, -0.2]) @ robot.fkm()

    # Use inverse kinematics to set the robot to the starting pose
    q = robot.ikm(htm0, q0=robot.q)
    robot.add_ani_frame(0, q)

    # Set desired pose
    htm_des = Utils.trn([0.5, 0, 0]) @ robot.fkm()
    desired_eef_frame.set_ani_frame(htm_des)

    # Simulation parameters
    time_max = 16  # 20
    dt = 0.01
    alpha = 1.5  # 3
    gamma = 0.1  # 0.5 0.3 0.1
    beta = 0.005
    h = 0.05  # 0.05
    qddotmax = 5

    # Initializations
    hist_dist = []
    hist_time = []
    hist_r = []
    hist_qddot = []
    hist_qdot = []
    hist_q = []
    hp = []
    dist_struct = None
    imax = round(time_max / dt)

    qdot = np.zeros((6,1))
    end_loop = False
    i = 0

    # Main loop
    while (not end_loop) and (i < imax):

        if i % 50 == 0 or i == imax - 1:
            sys.stdout.write('\r')
            sys.stdout.write("[%-20s] %d%%" % ('=' * round(20 * i / (imax - 1)), round(100 * i / (imax - 1))))
            sys.stdout.flush()

        # Compute the task function and task jacobian
        r, jac_r = robot.task_function(htm_des, np.array(q))
        # r[3:6] = np.sqrt(r[3:6])

        # Compute the task function and task jacobian in the next configuration
        r_next, jac_r_next = robot.task_function(htm_des, np.array(q + qdot * dt))
        # r_next[3:6] = np.sqrt(r_next[3:6])

        # Compute the time derivative of the jacobian and task function
        r_dot = (r_next - r) / dt
        jac_r_dot = (jac_r_next - jac_r) / dt

        # Create the h distance struct data
        dist_struct = robot.compute_dist(obstacle, h, h, np.array(q), old_dist_struct=dist_struct)
        dist_struct_next = robot.compute_dist(obstacle, h, h, np.array(q + qdot * dt), old_dist_struct=dist_struct)

        jac_dist = dist_struct.jac_dist_mat
        jac_dist_next = dist_struct_next.jac_dist_mat

        dist_vect = dist_struct.dist_vect
        dist_vect_next = dist_struct_next.dist_vect

        jac_dist_dot = (jac_dist_next - jac_dist) / dt
        dist_vect_dot = (dist_vect_next - dist_vect) / dt

        # Create the true distance struct data
        true_dist_struct = robot.compute_dist(obstacle, 0.0001, 0.0001, np.array(q), old_dist_struct=dist_struct)

        # Create the optimization problem and solve it for the joint acceleration
        a_hat = -Utils.dp_inv(jac_r, 0.001) @ (jac_r_dot @ qdot + alpha * r_dot)

        H = 2 * (np.transpose(jac_r) @ jac_r) + 2 * beta * np.identity(6)
        f = 2 * np.transpose(jac_r) @ (jac_r_dot @ qdot + 2 * alpha * r_dot + alpha * alpha * r) - 2 * beta * a_hat
        A = jac_dist
        b = -jac_dist_dot @ qdot - 2 * gamma * dist_vect_dot - gamma * gamma * dist_vect

        A = A[4, :]
        b = b[4, :]

        A = np.vstack((A, np.identity(6)))
        A = np.vstack((A, -np.identity(6)))

        m = np.shape(b)[0]

        b = np.vstack((b.reshape((m, 1)), -qddotmax * np.ones((6, 1))))
        b = np.vstack((b.reshape((m + 6, 1)), -qddotmax * np.ones((6, 1))))

        try:
            # u = quadprog.solve_qp(H, -f, np.transpose(A), b.reshape(m + 12, ))[0]
            qddot = solvers.qp(matrix(H), matrix(f), matrix(-A), matrix(-b.reshape(m + 12, )))['x']
            qddot = np.reshape(qddot, (6,1))

            # Integrate the joint acceleration
            q = q + qdot * dt
            qdot = qdot + qddot * dt
            robot.add_ani_frame(i * dt, q)


            # Record data for later plotting
            hist_dist.append(true_dist_struct[4].distance)
            hist_time.append((i - 1) * dt)
            hist_r.append(r.reshape((6,)))
            hist_q.append(q.reshape((6,)))
            hist_qdot.append(qdot.reshape((6,)))
            hist_qddot.append(qddot.reshape((6,)))
            hp.append(dist_struct[4].point_object)

            #print("time = "+str((i-1)*dt)+", q="+str(q.tolist()))

        except:
            end_loop = True

        i += 1

    # Run simulation
    sim.run()

    # Plot graphs

    Utils.plot(hist_time, hist_dist, "", "Time (s)", "True distance (m)", "dist")
    Utils.plot(hist_time, np.transpose(hist_q), "", "Time (s)", "Joint configuration (rad)", "q")
    Utils.plot(hist_time, np.transpose(hist_qdot), "", "Time (s)", "Joint speed (rad/s)", "qdot")
    Utils.plot(hist_time, np.transpose(hist_qddot), "", "Time (s)", "Joint acceleration (rad/s²)", "u")
    fig = Utils.plot(hist_time, np.transpose(hist_r), "", "Time (s)", "Task function",
                     ["posx", "posy", "posz", "orix", "oriy", "oriz"])

    return sim



from math import *
import numpy as np
from colour import Color
import plotly.express as px
import httplib2
import string as st
from scipy.linalg import null_space
from scipy.spatial import KDTree
from stl import mesh
import pywavefront as py_obj
import collada as py_col
from httplib2 import *


class Utils:
    """A library that contains some utilities for UAIbot. All of the functions are static."""

    #######################################
    # Constants
    #######################################

    _PI = 3.1415926
    _SQRTHALFPI = 1.2533141
    _SQRT2 = 1.4142135
    _CONSTJA = 2.7889
    _CONSTI0HAT1 = 0.24273
    _CONSTI0HAT2 = 0.43023

    UAIBOT_NAME_TYPES = ['uaibot.', 'cylinder.', 'box.', 'smoothbox.', 'ball.', 'robot.', 'simulation.', 'meshmaterial.',
                             'texture.',
                             'pointlight.', 'frame.', 'model3d.', 'links.', 'pointcloud.', 'vector.', 'rigidobject.',
                             '.group', '.htmldiv', '.Pedestrian', '.ObstacleColumn', '.ObstacleThinWall']

    IS_SIMPLE = ['uaibot.Ball', 'uaibot.Box', 'uaibot.Cylinder', 'uaibot.SmoothBox', 'uaibot.Cone']

    IS_GROUPABLE = ['uaibot.Ball', 'uaibot.Box', 'uaibot.SmoothBox', 'uaibot.Cylinder', 'uaibot.Frame',
                    'uaibot.RigidObject', 'uaibot.Group', 'uaibot.Robot', 'uaibot.PointLight', 'uaibot.Cone']

    IS_OBJ_SIM = ['uaibot.Ball', 'uaibot.Box', 'uaibot.SmoothBox', 'uaibot.Cylinder', 'uaibot.Cone', 'uaibot.Robot',
                  'uaibot.PointLight', 'uaibot.Frame', 'uaibot.PointCloud', 'uaibot.Vector',
                  'uaibot.RigidObject', 'uaibot.Group', 'uaibot.HTMLDiv', 'uaibot.Pedestrian', 'uaibot.ObstacleColumn', 'uaibot.ObstacleThinWall']

    #######################################
    # Basic functions
    #######################################

    @staticmethod
    def S(v):
        """
      Returns a 3x3 matrix that implements the cross product for a 3D vector  
      as a matricial product, that is, a matrix S(v) such that for any other 
      3D column  vector w, S(v)w = cross(v,w).
      
      Parameters
      ----------
      v : a 3D vector
          The vector for which the S matrix will be created.

      Returns
      -------
      S : 3x3 numpy matrix
          A matrix that implements the cross product with v.
      """
        vv = np.matrix(v).reshape((3,1))
        return np.matrix([[0, -vv[2,0], vv[1,0]],
                         [vv[2,0], 0, -vv[0,0]],
                         [-vv[1,0], vv[0,0], 0]])

    @staticmethod
    def rot(axis, angle):
        """
      Homogeneous transformation matrix that represents the rotation of an
      angle in an axis.
      
      Parameters
      ----------
      axis : a 3D vector
          The axis of rotation.
      
      angle: float
          The angle of rotation, in radians.

      Returns
      -------
      htm : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """
        a = np.reshape(axis, (3,))
        a = a / np.linalg.norm(a)
        K = Utils.S(a)
        Q = np.identity(3) + sin(angle) * K + (1 - cos(angle)) * (K @ K)
        return np.hstack([np.vstack([Q, np.matrix([0, 0, 0])]), np.matrix([[0], [0], [0], [1]])])

    @staticmethod
    def trn(vector):
        """
      Homogeneous transformation matrix that represents the displacement
      of a vector
      
      Parameters
      ----------
      vector : a 3D vector
          The displacement vector.
      
      Returns
      -------
      htm : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """
        v = np.matrix(vector, dtype=object).reshape((3,1))
        return np.matrix([[1, 0, 0, v[0,0]],
                         [0, 1, 0, v[1,0]],
                         [0, 0, 1, v[2,0]],
                         [0, 0, 0, 1]], dtype=object)

    @staticmethod
    def rotx(angle):
        """
      Homogeneous transformation matrix that represents the rotation of an
      angle in the 'x' axis.
      
      Parameters
      ----------
      angle: float
          The angle of rotation, in radians.

      Returns
      -------
      htm : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """
        return np.matrix([[1, 0, 0, 0],
                         [0, cos(angle), -sin(angle), 0],
                         [0, sin(angle), cos(angle), 0],
                         [0, 0, 0, 1]])

    @staticmethod
    def roty(angle):
        """
      Homogeneous transformation matrix that represents the rotation of an
      angle in the 'y' axis.
      
      Parameters
      ----------
      angle: float
          The angle of rotation, in radians.

      Returns
      -------
      htm : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """
        return np.matrix([[cos(angle), 0, sin(angle), 0],
                         [0, 1, 0, 0],
                         [-sin(angle), 0, cos(angle), 0],
                         [0, 0, 0, 1]])

    @staticmethod
    def rotz(angle):
        """
      Homogeneous transformation matrix that represents the rotation of an
      angle in the 'z' axis.
      
      Parameters
      ----------
      angle: float
          The angle of rotation, in radians.

      Returns
      -------
      htm : 4x4 numpy matrix
          The homogeneous transformation matrix.
      """
        return np.matrix([[cos(angle), -sin(angle), 0, 0],
                         [sin(angle), cos(angle), 0, 0],
                         [0, 0, 1, 0],
                         [0, 0, 0, 1]])

    @staticmethod
    def htm_rand(trn=1,rot=np.pi/2):
        """
      Returns a random homogeneous transformation matrix.

      Parameters
      ----------
      trn: float
          Maximum parameter for random translation in x, y and z.
          (default: 1)

      rot: float
          Maximum parameter for random rotation in x, y and z.
          (default: 1)

      Returns
      -------
      htm : 4x4 numpy matrix
          A homogeneous transformation matrix.
      """

        x = np.random.uniform(-trn / 2, trn / 2)
        y = np.random.uniform(-trn / 2, trn / 2)
        z = np.random.uniform(-trn / 2, trn / 2)
        ax = np.random.uniform(-rot / 2, rot / 2)
        ay = np.random.uniform(-rot / 2, rot / 2)
        az = np.random.uniform(-rot / 2, rot / 2)

        return Utils.trn([x,y,z]) * Utils.rotx(ax) * Utils.roty(ay) * Utils.rotz(az)

    @staticmethod
    def inv_htm(htm):
        """
      Given a homogeneous transformation matrix, compute its inverse.
      It is faster than using numpy.linalg.inv in the case of HTMs.
      
      Parameters
      ----------
      htm: 4X4 numpy array or nested list 
          Homogeneous transformation matrix of the rotation.

      Returns
      -------
      inv_htm: 4X4 numpy array
          The inverse of the transformation matrix.       
      """
        Q = htm[0:3, 0:3]
        p = htm[0:3, 3]

        inv_htm = np.matrix(np.zeros((4, 4)))
        inv_htm[0:3, 0:3] = Q.T
        inv_htm[0:3, 3] = -Q.T * p
        inv_htm[3, 3] = 1

        return inv_htm

    @staticmethod
    def axis_angle(htm):
        """
      Given an homogeneous transformation matrix representing a rotation, 
      return the rotation axis angle.
      
      Parameters
      ----------
      htm: 4X4 numpy array or nested list 
          Homogeneous transformation matrix of the rotation.

      Returns
      -------
      axis : 3D numpy vector
          The rotation axis.

      angle : float
          The rotation angle, in radians.        
      """
        Q = htm[0:3, 0:3]
        trace = Q[0, 0] + Q[1, 1] + Q[2, 2]
        angle = acos((trace - 1) / 2)
        G = Q * Q - 2 * cos(angle) * Q + np.identity(3)
        ok = False
        while not ok:
            v = np.matrix(np.random.uniform(-100, 100, size=(3,1)))
            w = np.matrix(np.random.uniform(-100, 100, size=(3,1)))
            r = G * v
            nr = np.linalg.norm(r)
            prod = w.T * r
            if nr > 0.01:
                ortr = w -  (r * prod) / (nr * nr)
                axis = Utils.S(ortr) * (Q * ortr)
                naxis = np.linalg.norm(axis)
                ok = naxis > 0.01

        axis = axis / naxis
        return axis, angle

    @staticmethod
    def euler_angles(htm):
        """
        Computer the Euler angles of a rotation matrix.
        Find alpha, beta and gamma such that.

        htm = Utils.rotz(alpha) * Utils.roty(beta) * Utils.rotx(gamma).

        Parameters
        ----------
        htm: 4X4 numpy array or nested list
            Homogeneous transformation matrix of the rotation.

        Returns
        -------
        alpha : float
            Rotation in z, in radians.
        beta : float
            Rotation in y, in radians.
        gamma : float
            Rotation in x, in radians.
        """

        Q = np.matrix(htm[0:3, 0:3])
        sy = sqrt(Q[0, 0] ** 2 + Q[1, 0] ** 2)

        if abs(sy) > 0.001:
            gamma = np.arctan2(Q[2, 1], Q[2, 2])
            beta = np.arctan2(-Q[2, 0], sy)
            alpha = np.arctan2(Q[1, 0], Q[0, 0])
        else:
            gamma = np.arctan2(-Q[1, 2], Q[1, 1])
            beta = np.arctan2(-Q[2, 0], sy)
            alpha = 0

        return alpha, beta, gamma


    @staticmethod
    def dp_inv(mat, eps = 0.001):
        """
      Compute the damped pseudoinverse of the matrix 'mat'.
      
      Parameters
      ----------
      mat: nxm numpy array
          The matrix to compute the damped pseudoinverse.
      
      eps: positive float
          The damping factor.
          (default: 0.001).

      Returns
      -------
      pinvA: mxn numpy array
          The damped pseudoinverse of 'mat'.
      """
        mat_int = np.matrix(mat)
        n = np.shape(mat_int)[1]
        return np.linalg.inv(mat_int.T * mat_int + eps * np.identity(n)) * mat_int.T

    @staticmethod
    def jac(f, x, delta=0.0001):
        """
      Compute the numerical Jacobian of a function f at the point x.
      Uses centralized finite difference to compute the derivatives.
      
      Parameters
      ----------
      f: function handle
          The function handle. It should accept a single 'm' dimensional numpy array/matrix
          which is a *column* matrix and return a single 'n' dimensional numpy/array/matrix
          which is a *column* matrix.

      x: m dimensional numpy *columsn* array/matrix
          Point in which the Jacobian will be computed. This object should be of the same nature of the input of
          f, so f(x) is well-defined.

      delta: float
          Variation used in the numerical differentiation
          (default: 0.0001)

      Returns
      -------
      jac: n x m numpy array
          The numerical Jacobian of f at point x. It is a n x m numpy array.
          If f_i(x) is the i-th entry of f(x) (1<=i<=n) and x_j the j-th variable (1<=j<=m), then
          the entry in the i-th row and j-th column is partial f_i/partial x_j evaluated at x.
      """

        if not (str(type(f)) == "<class 'function'>"):
            raise Exception("The parameter 'f' must be a function handle.")

        if not str(type(x)) == "<class 'numpy.ndarray'>" and not str(type(x)) == "<class 'numpy.matrix'>":
            raise Exception("Parameter 'x' should be a numpy array/matrix")

        if not(np.shape(x)[1] == 1):
            raise Exception("Parameter 'x' should be a column numpy array/matrix.")

        m = np.shape(x)[0]

        try:
            y = f(x)
            n = np.shape(y)[0]
        except:
            raise Exception("f(x) is not well defined.")

        jac = np.matrix(np.zeros((n, m)))
        idm = np.matrix(np.identity(m))

        for j in range(m):
            yp = f(x + idm[:, j] * delta)
            yn = f(x - idm[:, j] * delta)
            jac[:, j] = (yp - yn) / (2 * delta)

        return jac

    @staticmethod
    def hierarchical_solve(mat_a, mat_b, eps=0.001):
        """
      Solve the lexicographical unconstrained quadratic optimization problem

      lexmin_x ||mat_a[i]*x - b[i]||² + eps*||x||²

      with lower indexes having higher priority than higher indexes.

      Parameters
      ----------
      mat_a: A list of matrices (double arrays or numpy matrices).
          The matrices mat_a[i]. All must have the same number of columns.

      mat_b: A list of column vectors (double arrays or numpy matrices).
          The vectors mat_b[i]. The number of rows of mat_b[i] must be equal to the number
          of rows of mat_a[i].

      eps: positive float
          Damping parameter.
          (default: 0.001).

      Returns
      -------
      x: numpy column vector
          The solution x. For positive eps, the solution is always unique.
      """


        x_sol = Utils.dp_inv(mat_a[0], eps) * mat_b[0]

        if len(mat_a) > 1:

            null_mat_a = null_space(mat_a[0])

            if np.shape(null_mat_a)[1] > 0:
                mat_a_mod = []
                mat_b_mod = []
                for i in range(1, len(mat_a)):
                    mat_a_mod.append(mat_a[i] * null_mat_a)
                    mat_b_mod.append(mat_b[i] - mat_a[i] * x_sol)

                y_sol = Utils.hierarchical_solve(mat_a_mod, mat_b_mod, eps)
                return x_sol + null_mat_a * y_sol
            else:
                return x_sol
        else:
            return x_sol


    @staticmethod
    def interpolate(points):
        """
      Create a function handle that generates an one-time differentiable interpolated data from 'points'.

      The simplest case in when 'points' is a list with m elements. In this case, it will output a function f.
      When this function is evaluated at a scalar t, it will coincide with points[i] when t = i/m, that is,
      f(i/m) = points[i]. This function is once differentiable and periodic with period 1, so f(t+k)=f(t) for
      an integer k.

      The function can also use a n x m numpy array or lists as 'points'. In this case, f(t) is a n dimensional
      column vector in which its i-th entry is the same as computing f_i = interpolate(points[i]) and then
      computing f_i(t).

      Finally, t can be a list of k elements instead of just a scalar. In this case, f(t) is a n x k numpy matrix
      in which the element at row i and column j is the same as computing f_i = interpolate(points[i]) and then
      computing f_i(t[k]).


      Parameters
      ----------
      points: a n x m numpy array or lists
          Points to be interpolated.

      Returns
      -------
      f: function handle
          The function handle that implements the interpolation.
      """

        if not Utils.is_a_matrix(points):
            raise Exception("The parameter 'points' should be a n x m numpy array of numbers.")

        def aux_interpolate_single(arg_points):

            def int_aux_simp(arg_t, arg_n, arg_c):
                tn = arg_n * (arg_t % 1)
                ti = floor(tn)
                coef = arg_c[4 * ti: 4 * ti + 5]
                return coef[0] + coef[1] * tn + coef[2] * tn ** 2 + coef[3] * tn ** 3

            def int_aux(arg_t, arg_n, arg_c):
                return [int_aux_simp(tt, arg_n, arg_c) for tt in arg_t] if str(
                    type(arg_t)) == "<class 'list'>" else int_aux_simp(arg_t, arg_n, arg_c)

            n = len(arg_points)
            xn = np.array(arg_points).tolist()
            xn.append(arg_points[0])

            t = range(n + 1)
            A = np.zeros((3 * n, 4 * n))
            b = np.zeros((3 * n, 1))

            # Equality at initial points
            for p in range(n):
                A[p, 4 * p] = 1
                A[p, 4 * p + 1] = t[p]
                A[p, 4 * p + 2] = t[p] ** 2
                A[p, 4 * p + 3] = t[p] ** 3
                b[p] = xn[p]
            # Equality at final points
            for p in range(n):
                A[n + p, 4 * p] = 1
                A[n + p, 4 * p + 1] = t[p + 1]
                A[n + p, 4 * p + 2] = t[p + 1] ** 2
                A[n + p, 4 * p + 3] = t[p + 1] ** 3
                b[n + p] = xn[p + 1]
            # Equality of the derivative in the initial points
            for p in range(n):
                if not (p == n - 1):
                    A[2 * n + p, 4 * p] = 0
                    A[2 * n + p, 4 * p + 1] = 1
                    A[2 * n + p, 4 * p + 2] = 2 * t[p + 1]
                    A[2 * n + p, 4 * p + 3] = 3 * t[p + 1] ** 2
                    A[2 * n + p, 4 * (p + 1)] = -0
                    A[2 * n + p, 4 * (p + 1) + 1] = -1
                    A[2 * n + p, 4 * (p + 1) + 2] = -2 * t[p + 1]
                    A[2 * n + p, 4 * (p + 1) + 3] = -3 * t[p + 1] ** 2
                else:
                    A[2 * n + p, 4 * p] = 0
                    A[2 * n + p, 4 * p + 1] = 1
                    A[2 * n + p, 4 * p + 2] = 2 * t[p + 1]
                    A[2 * n + p, 4 * p + 3] = 3 * t[p + 1] ** 2
                    A[2 * n + p, 0] = -0
                    A[2 * n + p, 1] = -1
                    A[2 * n + p, 2] = 0
                    A[2 * n + p, 3] = 0

            # Create the objective function

            H = np.zeros((0, 0))
            for p in range(n):
                M = np.array([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 4, 6 * ((p + 1) ** 2 - p ** 2)],
                              [0, 0, 6 * ((p + 1) ** 2 - p ** 2), 12 * ((p + 1) ** 3 - p ** 3)]])
                nn = np.shape(H)[0]
                H = np.block([[H, np.zeros((nn, 4))], [np.zeros((4, nn)), M]])

            # Solve the optimization problem
            m1 = np.shape(A)[0]
            m2 = np.shape(A)[1]

            G = np.block([[H, np.transpose(A)], [A, np.zeros((m1, m1))]])
            g = np.block([[np.zeros((m2, 1))], [b]])
            y = np.linalg.solve(G, g)
            c = y[0: np.shape(H)[0]]
            c = c.reshape((1, np.shape(H)[0]))[0].tolist()

            # Create the function
            f = lambda ts: int_aux(ts, n, c)
            return f

        def aux_interpolate_multiple(arg_points, t):

            if not (Utils.is_a_vector(t) or Utils.is_a_number(t)):
                raise Exception(
                    "The parameter of the interpolation function must be either a number or a list of numbers.")

            y = np.zeros((0, len(t) if Utils.is_a_vector(t) else 1))
            for i in range(np.shape(arg_points)[0]):
                fun = aux_interpolate_single(arg_points[i])
                fun_out = fun(t)
                fun_out = np.array(fun_out).reshape((1, len(fun_out) if Utils.is_a_vector(fun_out) else 1))
                y = np.block([[y], [fun_out]])

            return np.matrix(y)

        return lambda t: aux_interpolate_multiple(points, t)

    #######################################
    # 3d objects manipulation
    #######################################

    @staticmethod
    def _dist_triangle_point(point, triangle):

        v0 = triangle[0,:].reshape((3,1))
        v1 = triangle[1,:].reshape((3,1))
        v2 = triangle[2,:].reshape((3,1))

        A = np.block([v1-v0,v2-v0])
        b = np.array(point).reshape((3,1))-v0

        #Create the solutions

        list_x = []

        list_x.append(np.linalg.inv(np.transpose(A) @ A) @ (np.transpose(A) @ b))

        prod0 = np.transpose(A[:, 0]) @ A[:, 0]
        prod1 = np.transpose(A[:, 1]) @ A[:, 1]

        val = float((np.transpose(A[:, 0]) @ b) / prod0)
        list_x.append(np.array([[val],[0]]))

        val = float((np.transpose(A[:, 1]) @ b) / prod1)
        list_x.append(np.array([[0], [val]]))

        val = float((np.transpose(A[:, 0]) @ (b-A[:, 1].reshape((3,1)))) / prod0)
        list_x.append(np.array([[val], [1]]))

        val = float((np.transpose(A[:, 1]) @ (b-A[:, 0].reshape((3,1)))) / prod1)
        list_x.append(np.array([[1], [val]]))

        list_x.append(np.array([[0],[0]]))
        list_x.append(np.array([[0],[1]]))
        list_x.append(np.array([[1],[0]]))
        list_x.append(np.array([[1],[1]]))

        val_min = float('inf')

        for i in range(9):
            if 1 >= list_x[i][0] >= 0 and 1 >= list_x[i][1]>=0:
                val_min = min(val_min,np.linalg.norm(A @ list_x[i] - b))

        return val_min

    @staticmethod
    def _distance_surface(point, mesh_triangle, kd_tree):

        PARAM_NO_KD_TREE = 30

        _, ind = kd_tree.query(point,PARAM_NO_KD_TREE)
        list_val = []
        for i in range(PARAM_NO_KD_TREE):
            list_val.append(Utils._dist_triangle_point(point,mesh_triangle[ind[i]]))

        val_min = min(list_val)
        return val_min

        #sum_exp=0
        #for i in range(PARAM_NO_KD_TREE):
        #    sum_exp+= exp(-(list_val[i]-val_min)/(h*h))
        #
        #return val_min - h * h * log(sum_exp/PARAM_NO_KD_TREE)

    @staticmethod
    def generate_connectivity_info(mesh_triangle, h):

        PARAM_NO_POINTS_GRAPH = 300
        PARAM_PERCENTAGE = 0.01

        #Compute the limits
        n = len(mesh_triangle)

        xmin = float('inf')
        ymin = float('inf')
        zmin = float('inf')
        xmax = -float('inf')
        ymax = -float('inf')
        zmax = -float('inf')

        triangle_center=[]

        for i in range(n):
            xmin = min(xmin,mesh_triangle[i][0,0],mesh_triangle[i][1,0],mesh_triangle[i][2,0])
            xmax = max(xmax,mesh_triangle[i][0,0],mesh_triangle[i][1,0],mesh_triangle[i][2,0])
            ymin = min(ymin,mesh_triangle[i][0,1],mesh_triangle[i][1,1],mesh_triangle[i][2,1])
            ymax = max(ymax,mesh_triangle[i][0,1],mesh_triangle[i][1,1],mesh_triangle[i][2,1])
            zmin = min(zmin,mesh_triangle[i][0,2],mesh_triangle[i][1,2],mesh_triangle[i][2,2])
            zmax = max(zmax,mesh_triangle[i][0,2],mesh_triangle[i][1,2],mesh_triangle[i][2,2])
            triangle_center.append( (mesh_triangle[i][0,:]+mesh_triangle[i][1,:]+mesh_triangle[i][2,:])/3)

        kd_tree = KDTree(triangle_center)
        dx = xmax - xmin
        dy = ymax - ymin
        dz = zmax - zmin
        limits = [xmin-0.15*dx, xmax+0.15*dx, ymin-0.15*dy, ymax+0.15*dy, zmin-0.15*dz, zmax+0.15*dz ]

        points_inside=[]
        points_outside =[]

        for i in range(2):
            for j in range(2):
                for k in range(2):
                    points_outside.append(np.array([limits[i], limits[2+j], limits[4+k]]))

        def gen_point():
            x = np.random.uniform(limits[0], limits[1])
            y = np.random.uniform(limits[2], limits[3])
            z = np.random.uniform(limits[4], limits[5])
            return np.array([x,y,z])


        def eval_fun(point):
            return Utils._distance_surface(point,mesh_triangle, kd_tree)

        EPSILON_DIST = PARAM_PERCENTAGE * max(dx,dy,dz)

        def line_is_free(point_1,point_2):
            dist_1 = eval_fun(point_1)
            if dist_1 <= EPSILON_DIST:
                return False

            dist_2 = eval_fun(point_2)
            if dist_2 <= EPSILON_DIST:
                return False

            dist_points = np.linalg.norm(point_1 - point_2)
            delta = (dist_1 - dist_2 + dist_points) / 2

            if dist_points < 0.5*EPSILON_DIST:
                return True

            if dist_1 - delta > 2*EPSILON_DIST:
                return True
            else:
                point_middle = (point_1+point_2)/2
                return line_is_free(point_1,point_middle) and line_is_free(point_middle, point_2)



        #Generate a list of points outside
        while len(points_outside) < PARAM_NO_POINTS_GRAPH/2:
            rand_point = gen_point()
            rand_point_outside = points_outside[np.random.randint(0,len(points_outside))]
            if line_is_free(rand_point,rand_point_outside):
                points_outside.append(rand_point)
                print(len(points_outside))

        #Generate a single point inside
        while len(points_inside)==0:
            rand_point = gen_point()
            if eval_fun(rand_point) >  EPSILON_DIST:
                connected_with_outside = False
                j = 0
                while not connected_with_outside and (j < len(points_outside)):
                    connected_with_outside = line_is_free(rand_point,points_outside[j])
                    j+= 1

                if not connected_with_outside:
                    points_inside.append(rand_point)

        print("Point inside found!")
        #Generate other points inside
        while len(points_inside) < PARAM_NO_POINTS_GRAPH/2:
            rand_point = gen_point()
            rand_point_inside = points_inside[np.random.randint(0,len(points_inside))]
            if line_is_free(rand_point,rand_point_inside):
                points_inside.append(rand_point)
                print(len(points_inside))
            else:
                print("Failed!")





        return kd_tree, limits, points_outside, points_inside

    @staticmethod
    def get_data_from_model(path):
        type = path[path.rfind(".") + 1:len(path) + 1]

        points = np.zeros((0,3))
        triangle = []

        if type=="obj":
            mesh_data = py_obj.Wavefront(path,encoding="iso-8859-1",parse=False,create_materials=True,collect_faces=True)
            mesh_data.parse()

            n = len(mesh_data.vertices)
            points = np.zeros((n, 3))
            for i in range(n):
                points[i, 0] = mesh_data.vertices[i][0]
                points[i, 1] = mesh_data.vertices[i][1]
                points[i, 2] = mesh_data.vertices[i][2]

            for i in range(len(mesh_data.mesh_list)):
                for j in range(len(mesh_data.mesh_list[i].faces)):
                    i1 = mesh_data.mesh_list[i].faces[j][0]
                    i2 = mesh_data.mesh_list[i].faces[j][1]
                    i3 = mesh_data.mesh_list[i].faces[j][2]
                    tri = np.block([ [points[i1,:]],  [points[i2,:]],  [points[i3,:]]])
                    triangle.append(tri)

        if type=="stl":

            mesh_data = mesh.Mesh.from_file(path)
            points = np.unique(mesh_data.vectors.reshape((int(mesh_data.vectors.size/3),3)), axis=0)
            triangle = mesh_data.vectors

        if type=="dae":

            mesh_data = py_col.Collada(path)
            for i in range(len(mesh_data.geometries)):
                for j in range(len(mesh_data.geometries[i].primitives[0])):
                    tri = mesh_data.geometries[i].primitives[0][j].vertices
                    triangle.append(tri)
                    points = np.block([[points],[tri]])

            points = np.unique(points,axis=0)

        return points, triangle

    #######################################
    # Type check functions
    #######################################

    @staticmethod
    def is_a_number(obj):
        """
      Check if the argument is a float or int number
      
      Parameters
      ----------
      obj: object
          Object to be verified.
      
      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """

        return str(type(obj)) in ["<class 'int'>", "<class 'float'>", "<class 'numpy.float64'>"]

    @staticmethod
    def is_a_natural_number(obj):
        """
      Check if the argument is a natural number (integer and >=0)
      
      Parameters
      ----------
      obj: object
          Object to be verified.
      
      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """

        return str(type(obj)) == "<class 'int'>" and obj >= 0

    @staticmethod
    def is_a_matrix(obj, n=None, m=None):
        """
      Check if the argument is a nxm matrix of floats.
      
      Parameters
      ----------
      obj: object
          Object to be verified.

      n: positive int
          Number of rows
          (default: it does not matter).

      m: positive int
          Number of columns
          (default: it does not matter).

      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """

        if str(type(obj)) == "<class 'numpy.ndarray'>" or str(type(obj)) == "<class 'numpy.matrix'>":
            return Utils.is_a_matrix(obj.tolist(), n, m)
        else:
            if str(type(obj)) == "<class 'list'>":

                is_all_a_num = True
                is_all_a_list = True

                for obj_el in obj:
                    is_all_a_num = is_all_a_num and Utils.is_a_number(obj_el)
                    is_all_a_list = is_all_a_list and (str(type(obj_el)) == "<class 'list'>")

                if is_all_a_num or is_all_a_list:

                    if is_all_a_num:
                        return ((m is None) or (m == len(obj))) and ((n == 1) or (n is None))

                    if is_all_a_list:
                        if (n is None) or (len(obj) == n):

                            m0 = len(obj[0])

                            if (m is None) or (m == m0):
                                is_correct = True
                                for obj_el in obj:
                                    is_correct = is_correct and len(obj_el) == m0
                                    if is_correct:
                                        for j in range(len(obj_el)):
                                            is_correct = is_correct and Utils.is_a_number(obj_el[j])

                                return is_correct
                            else:
                                return False
                        else:
                            return False
                else:
                    return False
            else:
                return False

    @staticmethod
    def is_a_vector(obj, n=None):
        """
      Check if the argument is a n vector of floats.
      
      Parameters
      ----------
      obj: object
          Object to be verified.

      n: positive int
          Number of elements
          (default: it does not matter).

      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """
        return Utils.is_a_matrix(obj, n, 1) or Utils.is_a_matrix(obj, 1, n)

    @staticmethod
    def is_a_pd_matrix(obj, n=None):
        """
      Check if the argument is a symmetric nxn positive (semi)-definite matrix.
      
      Parameters
      ----------
      obj: object
          Object to be verified.

      n: positive int
          Dimension of the square matrix
          (default: it does not matter).
    
      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """
        value = Utils.is_a_matrix(obj, n, n)

        if value:
            value = np.allclose(obj, np.transpose(obj), rtol=1e-05, atol=1e-08)

        if value:
            try:
                np.linalg.cholesky(obj)
            except:
                value = False

        return value

    @staticmethod
    def is_a_color(obj):
        """
      Check if the argument is a HTML-compatible string that represents a color.
      
      Parameters
      ----------
      obj: object
          Object to be verified.
      
      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """

        try:
            obj = obj.replace(" ", "")
            Color(obj)
            return True
        except:
            return False

    @staticmethod
    def is_a_name(string):
        """
      Check if the argument is a valid name for uaibot objects.
      Only characters [a-z], [A-z], [0-9] and '_' are allowed.
      However, variables should not begin with numbers.

      Parameters
      ----------
      string: string
          Name to be verified.

      Returns
      -------
      is_name: boolean
          If the name is a valid name.
      """

        if str(type(string)) == "<class 'str'>":
            allowed1 = set(st.ascii_lowercase + st.ascii_uppercase + st.digits + "_")
            allowed2 = set(st.ascii_lowercase + st.ascii_uppercase + "_")

            return (set(string) <= allowed1) and (set(string[0]) <= allowed2)
        else:
            return False

    @staticmethod
    def get_uaibot_type(obj):
        """
      Return the UAIBot type of the object.
      Return the empty string if it is not an UAIBot object.
      
      Parameters
      ----------
      obj: object
          Object to be verified.
      
      Returns
      -------
      obj_type: string
          UAIBot type.
      """
        type_str = str(type(obj))

        ind = -1
        k = 0
        while ind == -1 and k < len(Utils.UAIBOT_NAME_TYPES):
            ind = type_str.find(Utils.UAIBOT_NAME_TYPES[k])
            k += 1

        if ind == -1:
            return ""
        else:
            ind1 = type_str.rfind('.')
            ind2 = type_str.rfind('>')
            return "uaibot." + type_str[ind1 + 1:ind2 - 1]

    @staticmethod
    def is_a_simple_object(obj):
        """
      Check if the argument is a simple object.
      Check the constant 'Utils.IS_SIMPLE' for a list of simple objects.

      Parameters
      ----------
      obj: object
          Object to be verified.

      Returns
      -------
      is_type: boolean
          If the object is of the type.
      """
        return Utils.get_uaibot_type(obj) in Utils.IS_SIMPLE

    @staticmethod
    def is_a_groupable_object(obj):
        """
      Check if the argument is a groupable object.
      Check the constant 'Utils.IS_GROUPABLE' for a list of groupable objects.

      Parameters
      ----------
      obj: object
          Object to be verified.
      
      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """
        return Utils.get_uaibot_type(obj) in Utils.IS_GROUPABLE

    @staticmethod
    def is_a_obj_sim(obj):
        """
      Check if the argument is an object that can be put into the simulator.
      Check the constant 'Utils.IS_OBJ_SIM' for a list of objects that can be put in the simulator.
      
      Parameters
      ----------
      obj: object
          Object to be verified.
      
      Returns
      -------
      is_type: boolean
          If the object is of the type.   
      """
        return Utils.get_uaibot_type(obj) in Utils.IS_OBJ_SIM

    @staticmethod
    def is_url_available(url, types):
        """
      Try to access the content of the url 'url'. Also verifies if the content is one of the extensions contained in
      'types' (e.g, types = ['png', 'bmp', 'jpg', 'jpeg'] for images).

      Never throws an Exception, always returning a string with a message. Returns 'ok!' if and only if the
      url was succesfully acessed and has the correct file type.

      Parameters
      ----------
      url: string
          The url string.

      types: list of string
          The desired content extensions.

      Returns
      -------
      message: string
          Message.
      """

        if not (str(type(url)) == "<class 'str'>"):
            return " is not a valid url"

        ind = url.rfind('.')
        filetype = url[ind + 1:]

        if not (filetype in types):
            return " must be an file of the types: '" + str(types) + "'"

        try:
            h = httplib2.Http()
            resp = h.request(url, 'HEAD')
            if int(resp[0]['status']) < 400:
                return "ok!"
            else:
                return " : not able to access '" + url + "'."
        except:
            return " : not able to access '" + url + "'."

    #######################################
    # Plotting functions
    #######################################

    @staticmethod
    def plot(xv, yv, title="", xname="x", yname="y", labels=""):

        fig = px.line(width=800, height=400)

        # Error handling

        if not Utils.is_a_matrix(yv):
            raise Exception("The parameter 'yv' should be a matrix.")

        n = 1 if len(np.shape(yv)) == 1 else np.shape(yv)[0]
        m = np.shape(yv)[0] if len(np.shape(yv)) == 1 else np.shape(yv)[1]


        if xv is None:
            xv = list(np.arange(0, m))

        if Utils.is_a_number(xv):
            xv = list(np.arange(0, xv * m, xv))


        if not Utils.is_a_vector(xv):
            raise Exception("The parameter 'xv' should be a vector")

        if not Utils.is_a_matrix(yv, None, m):
            raise Exception("The parameter 'yv' should be a matrix with " + str(m) + " columns.")



        list_names = []

        if str(type(labels)) == "<class 'str'>":
            for i in range(n):
                list_names.append(labels + "_" + str(i + 1))
        else:
            if str(type(labels)) == "<class 'list'>" and len(labels) == n:
                for i in range(n):
                    if str(type(labels[i])) == "<class 'str'>":
                        list_names.append(labels[i])
                    else:
                        raise Exception(
                            "Optional parameter 'labels' must be either a string or a list of " + str(n) + " strings")
            else:
                raise Exception(
                    "Optional parameter 'labels' must be either a string or a list of " + str(n) + " strings")

        # end error handling
        if n > 1:
            for i in range(n):
                fig.add_scatter(x=xv, y=np.matrix(yv[i,:]).tolist()[0], mode="lines", name=list_names[i])
        else:
            fig.add_scatter(x=xv, y=np.matrix(yv).tolist()[0], mode="lines", name=list_names[0])

        fig.update_xaxes(title_text=xname)
        fig.update_yaxes(title_text=yname)
        fig.show()

        return fig

    #######################################
    # Distance computation functions
    #######################################

    @staticmethod
    def bissection(fun, t0, tf, eps):

        tA = t0
        fA = fun(tA)
        tB = tf
        fB = fun(tB)

        if abs(fA) < eps:
            return tA
        elif abs(fB) < eps:
            return tB
        else:
            while fA*fB>0:
                tA = 0.8*tA
                fA = fun(tA)
                tB = 1.2*tB
                fB = fun(tB)

            tm = (tA + tB) / 2
            fm = fun(tm)

            k=1

            while abs(fm)>eps:
                if fA*fm>0:
                   tA = tm
                   fA = fm
                else:
                   tB = tm
                   fB = fm

                tm = (tA + tB) / 2
                fm = fun(tm)
                k+=1


            return tm

    @staticmethod
    def cubicsolve(p,lam):
        q=abs(p)
        a = q/(2*lam)
        b = sqrt(a**2+1/(27*lam**3))
        u = np.sign(p)* ( (b+a)**(1/3) - (b-a)**(1/3))
        error = u-p+lam*(u**3)
        return u

    @staticmethod
    def fun_Int(v, h, L):

        def fun_J(u):
            return Utils._CONSTJA / ((Utils._CONSTJA - 1) * sqrt(Utils._PI * u * u) + sqrt(
                Utils._PI * u * u + Utils._CONSTJA * Utils._CONSTJA))

        v = abs(v)
        if v <= L:
            a1 = exp(-(L - v) ** 2 / (2 * h * h)) * fun_J((v - L) / (Utils._SQRT2 * h))
            a2 = exp(-(L + v) ** 2 / (2 * h * h)) * fun_J((v + L) / (Utils._SQRT2 * h))
            return -h * h * log(Utils._SQRTHALFPI * (h / (2 * L)) * (2 - a1 - a2))
        else:
            a1 = fun_J((v - L) / (Utils._SQRT2 * h))
            a2 = exp(-2 * L * v / (h * h)) * fun_J((v + L) / (Utils._SQRT2 * h))
            return 0.5 * (v - L) ** 2 - h * h * log(Utils._SQRTHALFPI * (h / (2 * L)) * (a1 - a2))

    @staticmethod
    def fun_Cir(v, h, r):

        def fun_I0hat(u):
            return pow(1 + 0.25 * u * u, -0.25) * (1 + Utils._CONSTI0HAT1 * u * u) / (1 + Utils._CONSTI0HAT2 * u * u)

        def fun_f(nu, rho):
            a1 = exp(-0.5 * (rho - nu) ** 2)
            a2 = exp(-0.5 * (rho + nu) ** 2)
            return rho * (a1 + a2) * fun_I0hat(rho * nu)

        def fun_fhat(nu, rho, arg_rhobar):
            a1 = exp(-0.5 * (rho - nu) ** 2 + 0.5 * (arg_rhobar - nu) ** 2)
            a2 = exp(-0.5 * (rho + nu) ** 2 + 0.5 * (arg_rhobar - nu) ** 2)
            return rho * (a1 + a2) * fun_I0hat(rho * nu)

        v = abs(v)
        N = 7
        node = [0.94910, -0.74153, -0.40584, 0, 0.40584, 0.74153, 0.94910]
        weight = [0.12948, 0.27970, 0.38183, 0.4179, 0.38183, 0.27970, 0.12948]

        if v <= r:
            f_low = max(0, sqrt((v / h) ** 2 + 1) - 3)
            f_up = min(r / h, sqrt((v / h) ** 2 + 1) + 3)
            delta = 0.5 * (f_up - f_low)
            y = 0
            for i in range(N):
                y = y + weight[i] * fun_f(v / h, f_low + delta * (node[i] + 1))

            y = delta * y
            return -h * h * log(y * (h / r) * (h / r))
        else:
            f_low = 0
            f_up = r / h
            delta = 0.5 * (f_up - f_low)
            rhobar = f_low + delta * (node[N - 1] + 1)
            y = 0
            for i in range(N):
                y = y + weight[i] * fun_fhat(v / h, f_low + delta * (node[i] + 1), rhobar)

            y = delta * y
            return 0.5 * (v - h * rhobar) ** 2 - h * h * log(y * (h / r) * (h / r))

    @staticmethod
    def fun_Sph(v, h, r):

        v = abs(v)
        c = 3 * (h ** 2) / (2 * r ** 3)

        if v <= r:
            if v == 0:
                return -h * h * log(
                    c * (-2 * r * exp(-(r * r) / (2 * h * h)) + 2 * r * exp(-Utils.fun_Int(0, h, r) / (h * h))))
            else:
                a1 = exp(-((r + v) * (r + v) / (2 * h * h)))
                a2 = exp(-((r - v) * (r - v) / (2 * h * h)))
                return -h * h * log(c * (h * h * (a1 - a2) / v + 2 * r * exp(-Utils.fun_Int(v, h, r) / (h * h))))

        else:
            a1 = exp(-(2 * r * v / (h * h)))
            a2 = 1
            return 0.5 * (v - r) * (v - r) - h * h * log(
                c * (h * h * (a1 - a2) / v + 2 * r * exp((0.5 * (v - r) * (v - r) - Utils.fun_Int(v, h, r)) / (h * h))))

    @staticmethod
    def softmin(x,h):
        minval = np.min(x)
        s=0

        for val in x:
            s+= exp(-h*(val-minval))

        return minval -(1/h)*np.log(s)

    @staticmethod
    def softselectmin(x, y, h):
        minval = np.min(x)
        s = 0

        coef = []
        for val in x:
            coef.append(exp(-(val - minval)/h))
            s += coef[-1]

        coef = [c/s for c in coef]

        sselect = 0 * y[0]
        for i in range(len(coef)):
            sselect += coef[i] * y[i]

        return sselect, minval - h * np.log(s/len(coef))


    @staticmethod
    def softmax(x, h):
        return Utils.softmin(x,-h)

    @staticmethod
    def softselectmax(x, y, h):
        return Utils.softselectmin(x,y,-h)

    @staticmethod
    def compute_aabbdist(obj1, obj2):

        w1, d1, h1 = obj1.aabb()
        w2, d2, h2 = obj2.aabb()
        delta = obj1.htm[0:3, -1] - obj2.htm[0:3, -1]

        dx = max(abs(delta[0, 0]) - (w1 + w2) / 2, 0)
        dy = max(abs(delta[1, 0]) - (d1 + d2) / 2, 0)
        dz = max(abs(delta[2, 0]) - (h1 + h2) / 2, 0)

        return np.sqrt(dx ** 2 + dy ** 2 + dz ** 2)

    @staticmethod
    def compute_dist(obj_a, obj_b, p_a_init=None, tol=0.001, no_iter_max=20):

        # Error handling

        if not Utils.is_a_simple_object(obj_a):
            raise Exception("The parameter 'obj_a' must be one of the following: " + str(Utils.IS_SIMPLE) + ".")

        if not Utils.is_a_simple_object(obj_b):
            raise Exception("The parameter 'obj_b' must be one of the following: " + str(Utils.IS_SIMPLE) + ".")

        if not (p_a_init is None or Utils.is_a_vector(p_a_init, 3)):
            raise Exception("The optional parameter 'p_a_init' must be a 3D vector or 'None'.")

        if not Utils.is_a_number(tol) or tol <= 0:
            raise Exception("The optional parameter 'tol' must be a nonnegative number.")

        if not Utils.is_a_natural_number(no_iter_max):
            raise Exception("The optional parameter 'no_iter_max' must be a nonnegative integer.")

        # end error handling

        if p_a_init is None:
            p_a = np.random.uniform(-3, 3, size=(3,))
        else:
            p_a = np.array(p_a_init)

        converged = False
        i = 0

        while (not converged) and i < no_iter_max:
            p_a_ant = p_a
            p_b, dp_a = obj_b.projection(p_a)
            p_a, dp_b = obj_a.projection(p_b)
            converged = np.linalg.norm(p_a - p_a_ant) < tol
            i += 1

        dist = np.linalg.norm(p_a - p_b)

        return p_a, p_b, dist



from utils import *
import numpy as np
from cvxopt import matrix, solvers
import math

def _const_control_mod(self, htm_des, q=None, htm=None, obstacles = [], dict_old_dist_struct=None,
                   eta_obs=0.5, eta_joint=0.5, eta_auto=0.5,
                   dist_safe_obs=0.01, dist_safe_auto=0.01,
                   max_dist_obs=np.inf, max_dist_auto = np.inf, task_rate_fun = 0.5, eps=0.001):


    solvers.options['show_progress'] = False
    np.set_printoptions(precision=4, suppress=True, linewidth=150)


    if q is None:
        q = self.q
    if htm is None:
        htm = np.matrix(self.htm)

    n = len(self._links)


    # Error handling
    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not str(type(obstacles)) == "<class 'list'>":
        raise Exception("The parameter 'list' should be a list of simple objects.")

    for obs in obstacles:
        if not Utils.is_a_simple_object(obs):
            raise Exception("The parameter 'list' should be a list of simple objects.")



    if not Utils.is_a_number(eta_joint) or eta_joint <= 0:
        raise Exception("The parameter 'eta_joint' should be a positive number.")

    if not Utils.is_a_number(eta_auto) or eta_auto <= 0:
        raise Exception("The parameter 'eta_auto' should be a positive number.")



    if not Utils.is_a_number(dist_safe_auto) or dist_safe_auto <= 0:
        raise Exception("The parameter 'dist_safe_auto' should be a positive number.")

    if not Utils.is_a_number(max_dist_obs) or max_dist_obs <= 0:
        raise Exception("The parameter 'max_dist_obs' should be a positive number.")

    if not Utils.is_a_number(max_dist_auto) or max_dist_auto <= 0:
        raise Exception("The parameter 'max_dist_auto' should be a positive number.")

    if not (Utils.is_a_number(task_rate_fun) and task_rate_fun > 0) and not str(type(task_rate_fun)) == "<class 'function'>":
        raise Exception("The parameter 'task_rate_fun' should be a positive number or a positive function.")

    if not Utils.is_a_number(eps) or eps <= 0:
        raise Exception("The parameter 'eps' should be a positive number.")

    # end error handling

    r, Jr = self.task_function(htm_des, q, htm)

    if str(type(task_rate_fun)) == "<class 'function'>":
        try:
            tg_dot_r = task_rate_fun(r)
            if Utils.is_a_matrix(tg_dot_r,6,1):
                Exception("The output of f should be a 6 dimensional column vector.")
        except:
            raise Exception("There is an error with function 'f'. Check its code.")
    else:
        tg_dot_r = task_rate_fun * r


    A = np.matrix(np.zeros(( 0, n )))
    b = np.matrix(np.zeros(( 0, 1 )))

    #Create constraint due to collision with obstacles
    if True:
        dict_dist_struct = {}

        for obs in obstacles:
            old_dist_struct = None if dict_old_dist_struct is None else dict_old_dist_struct[obs]
            dist_struct = self.compute_dist(obs, q, htm, old_dist_struct, max_dist=max_dist_obs)
            dict_dist_struct[obs] = dist_struct
            A = np.block([[A], [dist_struct.jac_dist_mat]])
            b = np.block([[b], [-eta_obs[obs] * (dist_struct.dist_vect - dist_safe_obs[obs])]])

    #Create joint limits constraints
    if not math.isinf(eta_joint):
        A = np.block([[A], [np.identity(n)]])
        b = np.block([[b], [-eta_joint * (q - self.joint_limit[:, 0])]])
        A = np.block([[A], [-np.identity(n)]])
        b = np.block([[b], [-eta_joint * (self.joint_limit[:, 1] - q)]])

    #Create auto collision constraints
    #TO DO

    #Create objective function

    H = 2 * Jr.T * Jr + 2 * eps*np.identity(n)
    f = 2 * Jr.T * tg_dot_r

    try:
        qdot = solvers.qp(matrix(H), matrix(f), matrix(-A), matrix(-b))['x']
        failure = False
    except:
        qdot = np.matrix(np.zeros( (n,1) ))
        failure=True

    return qdot, failure, r, dict_dist_struct, A, b


import os, sys

currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

from ._control_demo_1 import _control_demo_1
from ._control_demo_2 import _control_demo_2
from ._control_demo_3 import _control_demo_3


from ._lesson_demo_3 import _lesson_demo_3

class Demo:
    """
  A class that contains demonstrations in robotics.
  """

    #######################################
    # Control demos
    #######################################

    @staticmethod
    def control_demo_1():
        """
    Show a robotic manipulator drawing a circle in a drawboard.
    The task demands that the manipulator keep a constant pose while
    drawing.
    """
        return _control_demo_1()

    @staticmethod
    def control_demo_2():
        """
    Show two robotic manipulators cooperating to transport an object between
    two plates. Note that the relative pose between the two end-effectors must
    be kept.

    The control uses task priority framework using the null space of the task
    Jacobian. The task of keeping the relative pose is prioritized over the
    task of moving the plate.
    """
        return _control_demo_2()

    @staticmethod
    def control_demo_3():
        """
    Show a robot manipulator that must achieve a pose inside a hole. Collision must be avoided.

    Perform second order kinematic control with constraints that enforce
    collision avoidance.
    """
        return _control_demo_3()

    @staticmethod
    def lesson_demo_3(robot_creator=None, width=800, height=600):
        """
    Show a robot manipulator that must achieve a pose inside a hole. Collision must be avoided.

    Perform second order kinematic control with constraints that enforce
    collision avoidance.
    """
        return _lesson_demo_3(robot_creator, width, height)


from utils import *
import numpy as np
import math
from ._dist_struct_robot_obj import DistStructRobotObj


# Compute the distance from each link to an object, for the current configuration
# of the robot



def _compute_dist(self, obj, q=None, htm=None, old_dist_struct=None, tol=0.0005, no_iter_max=20,
                  max_dist = np.inf):
    n = len(self.links)

    if q is None:
        q = self.q

    if htm is None:
        htm = self.htm

    # Error handling
    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not Utils.is_a_simple_object(obj):
        raise Exception("The parameter 'obj' should be one of the following types: " + str(Utils.IS_SIMPLE) + ".")

    if not Utils.is_a_number(tol) or tol <= 0:
        raise Exception("The parameter 'tol' should be a positive number.")

    if not Utils.is_a_number(max_dist) or tol <= 0:
        raise Exception("The parameter 'max_dist' should be a positive number, or np.inf.")

    if not Utils.is_a_natural_number(no_iter_max) or no_iter_max <= 0:
        raise Exception("The parameter 'no_iter_max' should be a positive natural number.")

    if not (old_dist_struct is None):
        try:
            if not (id(old_dist_struct.obj) == id(obj) and id(old_dist_struct.robot) == id(self)):
                Exception("The parameter 'old_dist_struct' is a 'DistStructRobotObj' object, but it " \
                          "must have to be relative to the SAME robot object and SAME external object, and " \
                          "this is not the case.")
        except:
            raise Exception("The parameter 'old_dist_struct' must be a 'DistStructRobotObj' object.")

    # end error handling

    dist_struct = DistStructRobotObj(obj, self)

    jac_dh, mth_dh = self.jac_geo(q, "dh", htm)

    col_object_copy = []

    # Update all collision objects of all links
    for i in range(n):
        col_object_copy.append([])
        for j in range(len(self.links[i].col_objects)):
            temp_copy = self.links[i].col_objects[j][0].copy()
            htmd = self.links[i].col_objects[j][1]
            temp_copy.set_ani_frame(mth_dh[i][:, :] * htmd)
            col_object_copy[i].append(temp_copy)

    # Compute the distance structure

    for i in range(n):
        for j in range(len(self.links[i].col_objects)):

            est_dist = 0 if math.isinf(max_dist) else Utils.compute_aabbdist(obj, col_object_copy[i][j])

            if est_dist <= max_dist:

                if old_dist_struct is None:
                    p_obj_0 = np.matrix(np.random.uniform(-100, 100, size=(3, 1)))
                else:
                    try:
                        p_obj_0 = old_dist_struct.get_item(i, j).point_object
                    except:
                        p_obj_0 = np.matrix(np.random.uniform(-100, 100, size=(3, 1)))

                p_obj, p_obj_col, d = Utils.compute_dist(obj, col_object_copy[i][j], p_obj_0, tol, no_iter_max)

                jac_obj_col = jac_dh[i][0:3, :] - Utils.S(p_obj_col - mth_dh[i][0:3, 3]) * jac_dh[i][3:6, :]
                jac_dist = ((p_obj_col - p_obj).T * jac_obj_col) / d

                dist_struct._append(i, j, d, p_obj_col, p_obj, jac_dist)



    return dist_struct



import os, sys

currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

import numpy as np
from utils import *
from graphics.meshmaterial import *
from simobjects.group import *

from ._set_ani_frame import _set_ani_frame
from ._add_ani_frame import _add_ani_frame

from ._ikm import _ikm
from ._fkm import _fkm
from ._jac_geo import _jac_geo
from ._jac_jac_geo import _jac_jac_geo
from ._jac_ana import _jac_ana

from ._dyn_model import _dyn_model

from ._vector_field import _vector_field
from ._task_function import _task_function
from ._coop_task_function import _coop_task_function
from ._const_control import _const_control
from ._const_control_mod import _const_control_mod

from ._gen_code import _gen_code
from ._update_col_object import _update_col_object
from ._add_col_object import _add_col_object
from ._attach_object import _attach_object
from ._detach_object import _detach_object

from ._compute_dist import _compute_dist
from ._compute_dist_auto import _compute_dist_auto
from ._check_free_configuration import _check_free_configuration

from ._create_kuka_kr5 import _create_kuka_kr5
from ._create_epson_t6 import _create_epson_t6
from ._create_staubli_tx60 import _create_staubli_tx60
from ._create_kuka_lbr_iiwa import _create_kuka_lbr_iiwa
from ._create_abb_crb import _create_abb_crb
from ._create_darwin_mini import _create_darwin_mini
from ._create_kuka_kr5_per import _create_kuka_kr5_per
from ._create_franka_ermika_3 import _create_franka_ermika_3

#teste
#from robot._create_davinci import _create_davinci

class Robot:
    """
  A class that contains a robot object in uaibot.

  Parameters
  ----------

  name : string
      The robot name.
      (default: 'genRobot').

  htm_base_0 : 4x4 numpy array or 4x4 nested list
      The robot base's configuration.
      (default: 4x4 identity matrix).

  list_base_3d_obj : list of 'uaibot.Model3D' objects
      The list of 3d models of the base of the robot.
      If set to None, there is no base 3d object.
      (default: None).

  links : A list of 'uaibot.Link' objects
      The list of link objects.

  q0 : nd numpy vector or array
      The robot initial configuration.
      (default: zero vector).

  eef_frame_visible : boolean
      Set if the end-effector frame is visible.
      (default: True).

  joint_limit : n x 2 numpy array or None
      A n x 2 numpy array containing the joint limits, either in rad or meters.
      If set to 'None', use very large joint limits.
      (default: None).
  """

    #######################################
    # Attributes
    #######################################

    @property
    def q(self):
        """The current joint configuration."""
        return np.matrix(self._q)

    @property
    def q0(self):
        """The default joint configuration."""
        return np.matrix(self._q0)

    @property
    def htm(self):
        """
        The current base configuration in scenario coordinates.
        A 4x4 homogeneous matrix written is scenario coordinates.
        """
        return np.matrix(self._htm)

    @property
    def htm_base_0(self):
        """
        The constant homogeneous transformation between the HTM of the base and
        the HTM of the first Denavit-Hartenberg frame.
        """
        return np.matrix(self._htm_base_0)

    @property
    def htm_n_eef(self):
        """
        The constant homogeneous transformation between the HTM of the last
        Denavit-Hartenberg frame and the end-effector
        """
        return np.matrix(self._htm_n_eef)

    @property
    def links(self):
        """Data structures containing the links of the robot."""
        return self._links

    @property
    def attached_objects(self):
        """Data structures containing the objects attached into the robot."""
        return self._attached_objects

    @property
    def name(self):
        """Name of the object."""
        return self._name

    @property
    def list_object_3d_base(self):
        """The list of 3d objects that form the base."""
        return self._list_object_3d_base

    @property
    def eef_frame_visible(self):
        """If the frame attached to the end effector is visible"""
        return self._eef_frame_visible

    @property
    def joint_limit(self):
        """A n x 2 numpy array containing the joint limits, either in rad or meters"""
        return self._joint_limit


    #######################################
    # Constructor
    #######################################

    def __init__(self, name, links, list_base_3d_obj=None, htm=np.identity(4), htm_base_0=np.identity(4),
                 htm_n_eef = np.identity(4), q0=None, eef_frame_visible=True, joint_limits = None):
        # Error handling

        if not (Utils.is_a_name(name)):
            raise Exception(
                "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_matrix(htm_base_0, 4, 4):
            raise Exception("The parameter 'htm_base_0' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_matrix(htm_n_eef, 4, 4):
            raise Exception("The parameter 'htm_n_eef' should be a 4x4 homogeneous transformation matrix.")

        if not (str(type(links)) == "<class 'list'>"):
            raise Exception("The parameter 'links' should be a list of 'uaibot.Link' objects.")
        else:
            for link in links:
                if not Utils.get_uaibot_type(link) == "uaibot.Link":
                    raise Exception("The parameter 'links' should be a list of 'uaibot.Link' objects.")

        n = len(links)

        if not (q0 is None):
            self._q0 = np.matrix(q0).reshape((n,1))
        else:
            self._q0 = np.matrix(np.zeros((n,1)))

        if not (str(type(list_base_3d_obj)) == "<class 'list'>" or (list_base_3d_obj is None)):
            raise Exception("The parameter 'list_base_3d_obj' should be a list of 'uaibot.Model3D' objects.")
        else:
            for i in range(len(list_base_3d_obj)):
                if not (Utils.get_uaibot_type(list_base_3d_obj[i]) == "uaibot.Model3D"):
                    raise Exception(
                        "The parameter 'list_base_3d_obj' should be a list of 'uaibot.Model3D' objects.")

        if not Utils.is_a_vector(self._q0, n):
            raise Exception("The parameter 'q0' should be a " + str(n) + " dimensional vector.")

        if not str(type(eef_frame_visible)) == "<class 'bool'>":
            raise Exception("The parameter 'eef_frame_visible' must be a boolean.")

        if not joint_limits is None:
            self._joint_limit = joint_limits
        else:
            self._joint_limit = np.block([-10 * np.ones((n, 1)) , 10 * np.ones((n, 1))])

        if not Utils.is_a_matrix(self._joint_limit, n, 2):
            raise Exception("The parameter 'joint_limits' should be a " + str(n) + " x 2 numpy array.")

        for i in range(n):
            if self._joint_limit[i, 0]>self._joint_limit[i, 0]:
                raise Exception(
                    "In the parameter 'joint_limits', the minimum value (first column) must be smaller or equal than the maximum value (second column)")

        # end error handling

        self._frames = []
        self._list_object_3d_base = list_base_3d_obj
        self._htm = np.matrix(htm)
        self._name = name
        self._attached_objects = []
        self._links = links
        self._htm_base_0 = htm_base_0
        self._htm_n_eef = htm_n_eef
        self._eef_frame_visible = eef_frame_visible
        self._max_time = 0

        # Set initial total configuration
        self.set_ani_frame(self._q0, self._htm)

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):
        n = len(self.links)

        string = "Robot with name '" + self.name + "': \n\n"
        string += " Number of joints: " + str(n) + "\n"
        string += " Joint types: "

        for i in range(n):
            string += "R" if self._links[i].joint_type == 0 else "P"

        string += "\n"
        string += " Current configuration: " + str([round(num, 3) for num in np.ndarray.tolist(self.q)]) + "\n"
        string += " Current base HTM: \n" + str(self.htm) + "\n"
        string += " Current end-effector HTM: \n" + str(self.fkm())
        return string

    #######################################
    # Methods for configuration changing
    #######################################

    def add_ani_frame(self, time, q=None, htm=None, enforce_joint_limits = False):
        """
    Add a single configuration to the object's animation queue.

    Parameters
    ----------
    time: positive float
        The timestamp of the animation frame, in seconds.
    q : nd numpy vector or array
        The manipulator's joint configuration.
    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration.
        (default: the same as the current HTM).
    enforce_joint_limits: boolean
        If True and some q violates the joint limits (seen in the attribute 'joint_limit'), the
        respective q is clamped. Note that it DOES NOT issue any warning for this. So, be aware.
        (default: False).

    Returns
    -------
    None
    """
        return _add_ani_frame(self, time, q, htm, enforce_joint_limits)

    def set_ani_frame(self, q=None, htm=None, enforce_joint_limits = False):
        """
    Reset object's animation queue and add a single configuration to the 
    object's animation queue.

    Parameters
    ----------
    q : nd numpy vector or array
        The manipulator's joint configuration .
        (default: the current  joint configuration (robot.q) for the manipulator, q0).
    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration.
        (default: the same as the current HTM).
    enforce_joint_limits: boolean
        If True and some q violates the joint limits (seen in the attribute 'joint_limit'), the
        respective q is clamped. Note that it DOES NOT issue any warning for this. So, be aware.
        (default: False).
    Returns
    -------
    None
    """
        return _set_ani_frame(self, q, htm, enforce_joint_limits)

    #######################################
    # Methods for kinematics model
    #######################################

    def fkm(self, q=None, axis='eef', htm=None):
        """
    Compute the forward kinematics for an axis at a given joint and base
    configuration. Everything is written in the scenario coordinates.

    Parameters
    ----------
    q : nd numpy vector or array
        The manipulator's joint configuration.
        (default: the current  joint configuration (robot.q) for the manipulator).
    axis : string
        For which axis you want to compute the FK:
        'eef': for the end-effector;
        'dh': for all Denavit-Hartenberg axis;
        'com': for all center-of-mass axis.
        (default: 'eef').    
    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration.
        (default: the same as the current HTM).

    Returns
    -------
    htm_fk : 4x4 or nx4x4 numpy matrix
        For axis='eef', returns a single htm. For the other cases, return
        n htms as a nx4x4 numpy matrix.
    """
        return _fkm(self, q, axis, htm)

    def ikm(self, htm_target, q0=None, p_tol=0.005, a_tol=5, no_iter_max=2000, ignore_orientation=False):
        """
    Try to solve the inverse kinematic problem for the end-effector, given a
    desired homogeneous transformation matrix. It returns the manipulator
    configuration.

    Important: it disregards the current htm of the base of the robot. That is,
    it assumes that robot.htm = np.identity(4). You can easily consider other
    cases by transforming htm_target as Utils.inv_htm(robot.htm) * htm_target.

    Uses an iterative algorithm.

    The algorithm can fail, throwing an Exception when it happens.

    Parameters
    ----------
    htm_target : 4x4 numpy array or 4x4 nested list
        The target end-effector HTM, written in scenario coordinates.
    q0 : n-dimensional numpy vector or array
        Initial guess for the algorithm for the joint configuration.
        (default: a random joint configuration).
    p_tol : positive float
        The accepted error for the end-effector position, in meters.
        (default: 0.005 m).    
    a_tol : positive float
        The accepted error for the end-effector orientation, in degrees.
        (default: 5 degrees). 
    no_iter_max : positive int
        The maximum number of iterations for the algorithm.
        (default: 2000 iterations).
    ignore_orientation: boolean
        If True, the orientation part of the HTM is ignored. Task is position-only.
        (default: False)

    Returns
    -------
    q : n-dimensional numpy column vector
        The configuration that solves the IK problem.
    """
        return _ikm(self, htm_target, q0, p_tol, a_tol, no_iter_max, ignore_orientation)

    def jac_geo(self, q=None, axis='eef', htm=None):
        """
    Compute the geometric Jacobian for an axis at a given joint and base
    configuration. Also returns the forward kinematics as a by-product.
    Everything is written in the scenario coordinates.

    Parameters
    ----------
    q : nd numpy vector or array
        The manipulator's joint configuration 
        (default: the current  joint configuration (robot.q) for the manipulator).
    axis : string
        For which axis you want to compute the FK:
        'eef': for the end-effector;
        'dh': for all Denavit-Hartenberg axis;
        'com': for all center-of-mass axis.
        (default: 'eef').    
    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration.
        (default: the same as the current htm).

    Returns
    -------
    jac_geo : 6 x n or n x 6 x n numpy matrix
        For axis='eef', returns a single 6xn Jacobian. For the other cases, 
        return n Jacobians as a nx6xn numpy matrix.

    htm_out : 4 x 4 or n x 4 x 4 numpy matrix
        For axis='eef', returns a single htm. For the other cases, return
        n htms as a n x 4 x 4 numpy matrix.
    """
        return _jac_geo(self, q, axis, htm)

    def jac_ana(self, q=None, htm=None):
        """
    Compute the analytic Jacobian for the end-effector. The Euler angle
    convention is zyx. Also returns the end-effector htm and Euler angles
    as a by-product.

    Parameters
    ----------
    q : nd numpy vector or array
        The manipulator's joint configuration
        (default: the current  joint configuration (robot.q) for the manipulator).
    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration.
        (default: the same as the current htm).

    Returns
    -------
    jac_ana : 6 x n numpy matrix
        The analytic Jacobian.

    htm_eef : 4 x 4 numpy matrix
        The end-effector htm.

    phi : 3x1 numpy matrix
        The euler angles in the z (alpha), y (beta) and x (gamma) convention,
        as a column vector.
    """
        return _jac_ana(self, q, htm)

    def jac_jac_geo(self, q=None, axis='eef', htm=None):
        """
    Compute the Jacobians of the columns of the geometric Jacobian in the joint variable 'q'.
    This can be either to the end-effector frames (axis='eef'), to the Denavit-Hartenberg (DH) frames
    (axis='dh') or to the center-of-mass (COM) frames (axis='com').

    If axis ='dh' or 'com':

    If the robot has n links, jj_geo = robot.jac_jac_geo(q, htm) is a list of lists, in which
    jj_geo[i][j] is a 6 x n matrix that represents the Jacobian of the j-th column of the geometric
    Jacobian matrix of the i-th DH or COM frame.

    jj_geo[i][j] for j>i is not computed, because the j-th column of the geometric Jacobian matrix of the
    i-th DH or COM frame is always the 6 x n zero matrix, regardless of the 'q' and 'htm' chosen.

    jj_geo[i][j] could be alternatively computed numerically. For example, for axis='dh', by defining the function of q
    f = lambda q_var: np.matrix(robot.jac_geo(q = q_var, htm = htm, axis = 'dh')[0][i][j])
    and then computing the numerical Jacobian as Utils.jac(f,q).
    However, this function  is faster and has greater numerical accuracy, since it is computed analytically
    instead of numerically.

    If axis='eef', this is the same as computing axis='dh' but throwing away all but the last list away.


    Parameters
    ----------
    q : n-dimensional numpy vector or array
        The manipulator's joint configuration
        (default: the current  joint configuration (robot.q) for the manipulator).

    axis : string
        For which axis you want to compute the FK:
        'eef': for the end-effector;
        'dh': for all Denavit-Hartenberg axis;
        'com': for all center-of-mass axis.
        (default: 'eef').

    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration.
        (default: the same as the current htm).

    Returns
    -------
    jj_geo : list of lists of 6 x n numpy arrays (if axis='dh' or 'com') or a list of 6 x n numpy arrays (if axis=='eef')
        The Jacobian of the j-th column of the geometric Jacobian matrix of the i-th Denavit-Hartenberg frame.

    """
        return _jac_jac_geo(self, q, axis, htm)

    #######################################
    # Methods for dynamics model
    #######################################

    def dyn_model(self, q, qdot):
        """
    Compute the three robot's dynamic model terms, in a given joint configuration 'q'
    and joint speed 'qdot'.

    Parameters
    ----------
    q : n-dimensional numpy vector or array
        The manipulator's joint configuration.

    qdot : n-dimensional numpy vector or array
        The manipulator's joint configuration speed.

    Returns
    -------
    dyn_m : n x n numpy array
        The generalized inertia matrix at the joint configuration q.

    dyn_c : n-dimensional numpy column vector
        The generalized Coriolis-Centrifugal torques at the joint
        configuration q and joint configuration speed qdot.

    dyn_g : n-dimensional numpy column vector
        The generalized gravity torques at the joint configuration q.

    """
        return _dyn_model(self, q, qdot)

    #######################################
    # Methods for control
    #######################################
    @staticmethod
    def vector_field(curve, alpha=1, const_vel=1):
        """
    Computes a handle to a vector field function fun(p). Uses the vector field
    presented in 
    
    "Adriano M. C. Rezende; Vinicius M. Goncalves; Luciano C. A. Pimenta: 
    Constructive Time-Varying Vector Fields for Robot Navigation 
    IEEE Transactions on Robotics (2021)". 
    
    The vector field has constant velocity and use the function 
    G(u) = (2/pi)*atan(alpha*u).


    Parameters
    ----------
    curve : nxm numpy array or nxm nested list
        Curve, described as sampled points. Each one of the n rows should 
        contain a m-dimensional float vector that is the n-th m-dimensional
        sampled point of the curve. 
 
    alpha : positive float
        Controls the vector field behaviour. Greater alpha's imply more 
        robustness to the vector field, but increased velocity and acceleration
        behaviours. Used in G(u) = (2/pi)*atan(alpha*u)
        (default: 1).

    const_vel : positive float
        The constant velocity of the vector field. The signal of this number 
        controls the direction of rotation 
        (default: 1).

    Returns
    -------
    fun: a function handle that you can call as f(p), returning a numpy column vector. 'p' should be a
         m-dimensional vector.
    """

        return _vector_field(curve, alpha, const_vel)

    def task_function(self, htm_des, q=None, htm=None):
        """
    Computes the 6-dimensional task function for end-effector pose control,  
    given a joint configuration, a base configuration and the desired pose 
    'htm_des'.

    The first three entries are position error, and the three last entries are
    orientation error.

    Everything is written in scenario coordinates. 

    Also returns the Jacobian of this function.

    Parameters
    ----------
    htm_des : 4x4 numpy array or 4x4 nested list
        The desired end-effector pose. 
 
    q : nd numpy vector or array
        The manipulator's joint configuration.
        (default: the current  joint configuration (robot.q) for the manipulator).

    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration.
        (default: the same as the current htm).

    Returns
    -------
    r : 6-dimensional numpy column vector
        The task function.

    jac_r : 6 x n numpy matrix
        The respective task Jacobian.
    """
        return _task_function(self, htm_des, q, htm)

    @staticmethod
    def coop_task_function(robot_a, robot_b, htm_a_des, htm_a_b_des, q_a=None, q_b=None):
        """
    Computes the 12-dimensional task function for end-effector pose control
    of two robots 'robot_a', 'robot_b' given their respective configurations
    q_a and q_b.

    
    The first three components are relative position error.  

    The second three components are relative orientation error.  

    The third three components are position error for 'robot_a'.

    The third three components are orientation error for 'robot_a'.
    
    Everything is written in scenario coordinates. 
    
    Also returns the Jacobian of this function.

    Parameters
    ----------
    robot_a :robot object
        The first robot.

    robot_b :robot object
        The second robot.

    htm_a_des :4x4 numpy array or 4x4 nested list
        The desired pose for the end-effector of 'robot_a'.

    htm_a_b_des :4x4 numpy array or 4x4 nested list
        The desired relative pose between the end-effector of 'robot_a' and
        'robot_b'. That is, inv(htmA) * htmB.

    q_a : nd numpy vector or array
        'robot_a'' joint configuration
        (default: the current  joint configuration (robot.q) for 'robot_a').

    q_b : md numpy vector or array
        'robot_b'' joint configuration
        (default: the current  joint configuration (robot.q) for 'robot_b').

    Returns
    -------
    r : 12-dimensional numpy column vector
        The task vector.

    jac_r : 12 x (n+m) numpy matrix
        The respective task Jacobian.
    """
        return _coop_task_function(robot_a, robot_b, htm_a_des, htm_a_b_des, q_a, q_b)

    def const_control(self, htm_des, q=None, htm=None, obstacles = [], dict_old_dist_struct=None,
                   eta_obs=0.5, eta_joint=0.5, eta_auto=0.5,
                   dist_safe_obs=0.01, dist_safe_auto=0.01,
                   max_dist_obs=np.inf, max_dist_auto = np.inf, task_rate_fun = 0.5, eps=0.001):
        """
    Computes the constrained joint velocity aiming a target pose (htm_des).
    Consider three kind of constraints: collision with obstacles, joint limits and
    auto collision.

    The obstacles should be simple objects (see Utils.IS_SIMPLE for the list).

    Parameters
    ----------

    htm_des :4x4 numpy array or 4x4 nested list
        The desired pose for the end-effector of the robot.

    q : nd numpy vector or array
        The joint configuration in which we want to compute the control action.
        (default: the current joint configuration for the robot).

    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration.
        (default: the same as the current HTM).

    obstacles : list of simple objects
        A list of obstacles as simple objects (see Utils.IS_SIMPLE)
        (default: empty list).

    dict_old_dist_struct: a dictionary that maps each obstacle to a 'DistStructRobotObj'
        Used to speed up computation using data from a previous run. The output
        'dict_dist_struct' of this function can be fed as this parameter in a next
        call. If None is provided, this speed up is ignored.
        (default: None)

    eta_obs : positive float
        Parameter that controls obstacle avoidance, measured in s^(-1)
        (inverse seconds). The rule is that the rate of approximation
        to each obstacle, in meters per seconds, should be at most:

        eta_obs * (distance-safe_distance).

        So, the higher is this value, the less the robot
        is "afraid" of obstacles. Not that if this value is too high
        there can be collisions. If this value is np.inf, obstacle avoidance
        is skipped.
        (default: 0.5 s^(-1)).

    eta_joint : positive float
        Parameter that controls joint limits avoidance, measured in s^(-1)
        (inverse seconds). The rule is that the rate of approximation
        to each joint limit, in meters per seconds, should be at most:

        eta_joint * (distance to joint limit).

        So, the higher is this value, the less the robot
        is "afraid" of joint limits. Not that if this value is too high
        there can be collisions. If this value is np.inf, joint limits
        are skipped.
        (default: 0.5 s^(-1)).

    eta_joint : positive float
        Parameter that controls joint limits avoidance, measured in s^(-1)
        (inverse seconds). The rule is that the rate of approximation
        to each joint limit, in meters per seconds, should be at most:

        eta_joint * (distance to joint limit -safe_distance).

        So, the higher is this value, the less the robot
        is "afraid" of joint limits. Not that if this value is too high
        there can be collisions. If this value is np.inf, joint limits
        are skipped.
        (default: 0.5 s^(-1)).

    eta_auto : positive float
        Parameter that controls auto collision avoidance, measured in s^(-1)
        (inverse seconds). The rule is that the rate of approximation
        between each non-sequential link, in meters per seconds, should be at most:

        eta_auto * (distance -safe_distance).

        So, the higher is this value, the less the robot
        is "afraid" of auto collision. Not that if this value is too high
        there can be collisions. If this value is np.inf, auto collision is
        skipped.
        (default: 0.5 s^(-1)).

    dist_safe_obs : positive float
        Parameter that controls a safety margin to collision to obstacles,
        in meters.
        (default: 0.01 meter).

    dist_safe_auto : positive float
        Parameter that controls a safety margin to auto collision,
        in meters.
        (default: 0.01 meter).

    max_dist_obs : positive float
        Controls the maximum distance considered to skip collision between the object
        and obstacles. See the parameter 'max_dist' in the function 'compute_dist' of
        the robot.
        (default: np.inf).

    max_dist_auto : positive float
        Controls the maximum distance considered when skipping collision between the object
        and itself (auto collision). See the parameter 'max_dist' in the function 'compute_dist' of
        the robot.
        (default: np.inf).

    task_rate_fun : positive float or a function handle
        const_fun should be either a function handle that receives a 6x1 numpy column vector and
        outputs a 6x1 numpy column vector or a positive float.

        If task_rate_fun is a function f(r), it controls the target task function decrease. So we
        aim to have the task function decrease dr/dt as -f(r). In order to be suitable function, there
        should exist a positive definite function V(r) such that grad -V(r).T * f(r) <=0. This is
        not checked and is a responability of the user.

        If task_rate_fun is a positive float, is the same result as using the function handle
        lambda r : task_rate_fun*r. In this case V(r) = ||r||² is suitable.
        (default: 0.5).

    eps : positive float
        Damping factor. Control inputs generated by this procedure can be quite noisy/discontinuous. The
        higher is this value, the less is this behaviour. However, this can generate steady state errors
        in the controller.
        (default: 0.001).

    Returns
    -------

    qdot, failure, r, dict_dist_struct

    qdot : n-dimensional numpy column vector
        The joint velocity that can be used to control.

    failure: boolean
        The controller is computed using a quadratic program. If the problem is unfeasible,
        then this variable returns 'false. This typically happens when the configuration
        'q' violates one of the constraints (collision to obstacle, joint limits or
        auto collision).

    r : 6-dimensional numpy column vector
        The task vector.

    dict_dist_struct : a dictionary that maps each obstacle to a 'DistStructRobotObj'
        Used to speed up computation using data from a previous run. This output can be
        fed as the parameter 'dict_old_dist_struct' in a next call.
    """

        return _const_control(self, htm_des, q, htm, obstacles, dict_old_dist_struct,
                              eta_obs, eta_joint, eta_auto, dist_safe_obs, dist_safe_auto,
                              max_dist_obs, max_dist_auto, task_rate_fun, eps)

    def const_control_mod(self, htm_des, q=None, htm=None, obstacles = [], dict_old_dist_struct=None,
                   eta_obs=0.5, eta_joint=0.5, eta_auto=0.5,
                   dist_safe_obs=0.01, dist_safe_auto=0.01,
                   max_dist_obs=np.inf, max_dist_auto = np.inf, task_rate_fun = 0.5, eps=0.001):


        return _const_control_mod(self, htm_des, q, htm, obstacles, dict_old_dist_struct,
                              eta_obs, eta_joint, eta_auto, dist_safe_obs, dist_safe_auto,
                              max_dist_obs, max_dist_auto, task_rate_fun, eps)
    #######################################
    # Methods for simulation
    #######################################

    def gen_code(self):
        """Generate code for injection."""
        return _gen_code(self)

    def update_col_object(self, time):
        """
        Update internally the objects that compose the collision model to the
        current configuration of the robot.
        """
        _update_col_object(self, time)

    def add_col_object(self, sim):
        """
        Add the objects that compose the collision model to a simulation.

        Parameters
        ----------
        sim : 'Simulation' object
            'Simulation' object.
    """
        _add_col_object(self, sim)

    def attach_object(self, obj):
        """
        Attach an object to the end-effector.
        The list of the type of objects that can be grouped can be seen in 'Utils.IS_GROUPABLE'.

        Parameters
        ----------
        obj : Object that is one of the types in 'Utils.IS_GROUPABLE'
            Object to be attached.
    """
        _attach_object(self, obj)

    def detach_object(self, obj):
        """
        Detach an object (ball, box or cylinder) to the end-effector.

        Parameters
        ----------
        obj : object
            Object to be detached.
    """
        _detach_object(self, obj)

    def set_htm_to_eef(self, htm):
        self._htm_n_eef = htm

    #######################################
    # Robot constructors
    #######################################

    @staticmethod
    def create_kuka_kr5(htm=np.identity(4), name='kukakr5', color="#df6c25", opacity=1, eef_frame_visible=True):
        """
    Create a Kuka KR5 R850, a six-degree of freedom manipulator.
    Thanks Sugi-Tjiu for the 3d model (see https://grabcad.com/library/kuka-kr-5-r850).

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The initial base configuration for the robot.
        (default: np.identity(4))
 
    name : string
        The robot name.
        (default: 'kukakr5').

    htm : color
        A HTML-compatible string representing the object color.
        (default: '#df6c25')'.

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Robot object
        The robot.

    """
        base_3d_obj, links, htm_base_0, htm_n_eef, q0, joint_limits = _create_kuka_kr5(htm, name, color, opacity)
        return Robot(name, links, base_3d_obj, htm, htm_base_0, htm_n_eef, q0, eef_frame_visible, joint_limits)

    @staticmethod
    def create_kuka_kr5_per(htm=np.identity(4), name='kukakr5', color="#df6c25", opacity=1, eef_frame_visible=True, per=0.05):
        """
    Create a Kuka KR5 R850, a six-degree of freedom manipulator.
    Thanks Sugi-Tjiu for the 3d model (see https://grabcad.com/library/kuka-kr-5-r850).

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The initial base configuration for the robot.
        (default: np.identity(4))

    name : string
        The robot name.
        (default: 'kukakr5').

    htm : color
        A HTML-compatible string representing the object color.
        (default: '#df6c25')'.

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Robot object
        The robot.

    """
        base_3d_obj, links, htm_base_0, htm_n_eef, q0, joint_limits = _create_kuka_kr5_per(htm, name, color, opacity, per)
        return Robot(name, links, base_3d_obj, htm, htm_base_0, htm_n_eef, q0, eef_frame_visible, joint_limits)

    @staticmethod
    def create_epson_t6(htm=np.identity(4), name='epsont6', color="white", opacity=1, eef_frame_visible=True):
        """
    Create an Epson T6, a SCARA manipulator.
    Thanks KUA for the 3d model (see https://grabcad.com/library/epson-t6-scara-robot-1).

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The initial base configuration for the robot.
        (default: np.identity(4))

    name : string
        The robot name.
        (default: 'epsont6').

    htm : color
        A HTML-compatible string representing the object color.
        (default: 'white')'.

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Robot object
        The robot.

    """
        base_3d_obj, links, htm_base_0, htm_n_eef, q0, joint_limits = _create_epson_t6(htm, name, color, opacity)
        return Robot(name, links, base_3d_obj, htm, htm_base_0, htm_n_eef, q0, eef_frame_visible, joint_limits)

    @staticmethod
    def create_staubli_tx60(htm=np.identity(4), name='staublitx60', color="#ff9b00", opacity=1, eef_frame_visible=True):
        """
    Create a Staubli TX60, a six degree of freedom manipulator.
    Model taken from the ROS github repository (https://github.com/ros-industrial/staubli).

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The initial base configuration for the robot.
        (default: np.identity(4))

    name : string
        The robot name.
        (default: 'staublitx60').

    htm : color
        A HTML-compatible string representing the object color.
        (default: '#ff9b00')'.

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Robot object
        The robot.

    """
        base_3d_obj, links, htm_base_0, htm_n_eef, q0, joint_limits = _create_staubli_tx60(htm, name, color, opacity)
        return Robot(name, links, base_3d_obj, htm, htm_base_0, htm_n_eef, q0, eef_frame_visible, joint_limits)

    @staticmethod
    def create_kuka_lbr_iiwa(htm=np.identity(4), name='kukalbriiwa', color="silver", opacity=1, eef_frame_visible=True):
        """
    Create a Kuka LBR IIWA 14 R820, a seven degree of freedom manipulator.
    Model taken from the ROS github repository (https://github.com/ros-industrial/kuka_experimental).

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The initial base configuration for the robot.
        (default: np.identity(4))

    name : string
        The robot name.
        (default: 'kukalbriiwa').

    htm : color
        A HTML-compatible string representing the object color.
        (default: 'silver')'.

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Robot object
        The robot.

    """
        base_3d_obj, links, htm_base_0, htm_n_eef, q0, joint_limits = _create_kuka_lbr_iiwa(htm, name, color, opacity)
        return Robot(name, links, base_3d_obj, htm, htm_base_0, htm_n_eef, q0, eef_frame_visible, joint_limits)

    @staticmethod
    def create_franka_ermika_3(htm=np.identity(4), name='frankaermika', color="silver", opacity=1, eef_frame_visible=True):
        """
    Create a Franka Ermika 3, a seven degree of freedom manipulator.
    Model taken from the ROS github repository (https://github.com/BolunDai0216/FR3Env/tree/d5218531471cadafd395428f8c2033f6feeb3555/FR3Env/robots/meshes/visual).

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The initial base configuration for the robot.
        (default: np.identity(4))

    name : string
        The robot name.
        (default: 'frankaermika').

    htm : color
        A HTML-compatible string representing the object color.
        (default: 'silver')'.

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Robot object
        The robot.

    """
        base_3d_obj, links, htm_base_0, htm_n_eef, q0, joint_limits = _create_franka_ermika_3(htm, name, color, opacity)
        return Robot(name, links, base_3d_obj, htm, htm_base_0, htm_n_eef, q0, eef_frame_visible, joint_limits)

    @staticmethod
    def create_abb_crb(htm=np.identity(4), name='abbcrb', color="white", opacity=1, eef_frame_visible=True):
        """
    Create a ABB CRB 15000, a six degree of freedom manipulator.
    Model taken from the ROS github repository (https://github.com/ros-industrial/abb_experimental).

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The initial base configuration for the robot.
        (default: np.identity(4))

    name : string
        The robot name.
        (default: 'abbcrb').

    htm : color
        A HTML-compatible string representing the object color.
        (default: 'white')'.

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Robot object
        The robot.

    """
        base_3d_obj, links, htm_base_0, htm_n_eef, q0, joint_limits = _create_abb_crb(htm, name, color, opacity)
        return Robot(name, links, base_3d_obj, htm, htm_base_0, htm_n_eef, q0, eef_frame_visible, joint_limits)

    @staticmethod
    def create_darwin_mini(htm=np.identity(4), name="darwin_mini", color="#3e3f42", opacity=1, eef_frame_visible=True):
        """
    Create an (oversized) Darwin Mini, a humanoid robot.
    Thanks to Alexandre Le Falher for the 3D model (https://grabcad.com/library/darwin-mini-1).

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The initial base configuration for the robot.
        (default: np.identity(4))

    name : string
        The robot name.
        (default: 'darwin_mini').

    htm : color
        A HTML-compatible string representing the object color.
        (default: '#3e3f42').

    opacity : positive float between 0 and 1
        The opacity of the robot. 1 = fully opaque and 0 = transparent.
        (default: 1)

    Returns
    -------
    robot : Group object
        The robot. It is composed of a group of six objects: the two arms and legs (members of 'Robot' class)
        and the chest and head (both 'RigidObject' class)


    """

        param_arm_left, param_arm_right, param_leg_left, param_leg_right, head, chest = _create_darwin_mini(htm, name,
                                                                                                            color,
                                                                                                            opacity)
        desl_z = htm * Utils.trn([0, 0, -0.18])
        robot_arm_left = Robot(name + "__arm_left", param_arm_left[1], param_arm_left[0],
                               desl_z * Utils.trn([0, 0.14, 1]) * Utils.rotx(-3.14 / 2), param_arm_left[2],
                               param_arm_left[3], [np.pi/2,0.3,0], eef_frame_visible,
                               param_arm_left[5])

        robot_arm_right = Robot(name + "__arm_right", param_arm_right[1], param_arm_right[0],
                                desl_z * Utils.rotz(3.14) * Utils.trn([0, 0.14, 1]) * Utils.rotx(-3.14 / 2),
                                param_arm_right[2],
                                param_arm_right[3], [np.pi/2,0.3,0], eef_frame_visible, param_arm_right[5])

        robot_leg_left = Robot(name + "__leg_left", param_leg_left[1], param_leg_left[0],
                               desl_z * Utils.trn([0, -0.1, 0.7]) * Utils.roty(np.pi / 2) * Utils.rotz(-np.pi / 2),
                               param_leg_left[2], param_leg_left[3], [np.pi/2, 0, 0, np.pi/2], eef_frame_visible, param_leg_left[5])

        robot_leg_right = Robot(name + "__leg_right", param_leg_right[1], param_leg_right[0],
                                desl_z * Utils.trn([0, 0.1, 0.7]) * Utils.roty(np.pi / 2) * Utils.rotz(-np.pi / 2),
                                param_leg_right[2], param_leg_right[3], [np.pi/2, 0, 0, np.pi/2], eef_frame_visible, param_leg_right[5])

        return Group([robot_arm_left, robot_arm_right, robot_leg_left, robot_leg_right, head, chest])

    #######################################
    # Distance computation and collision
    #######################################

    def compute_dist(self, obj, q=None, htm=None, old_dist_struct=None, tol=0.0005,
                     no_iter_max=20, max_dist = np.inf):
        """
    Compute the  distance structure from each one of the robot's link to a
    'simple' external object (see Utils.IS_SIMPLE), given a joint and base
    configuration.

    This function can be faster if some distance computations are avoided.
    See the description of the parameter 'max_dist'.

    Use an iterative algorithm, based on projections
    (Von Neumann's cyclic projection algorithm).

    Parameters
    ----------
    obj : a simple object (see Utils.IS_SIMPLE)
        The external object for which the distance structure is going to be 
        computed, for each robot link.

    q : nd numpy vector or array
        The manipulator's joint configuration.
        (default: the current  joint configuration (robot.q) for the manipulator).

    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration.
        (default: the same as the current htm).

    old_dist_struct : 'DistStructRobotObj' object
        'DistStructRobotObj' obtained previously for the same robot and external object.
        Can be used to enhance the algorithm speed using the previous closest
        point as an initial guess.
        (default: None).

    tol : positive float
        Tolerance for convergence in the iterative algorithm, in meters.
        (default: 0.0005 m).

    no_iter_max : positive int
        The maximum number of iterations for the algorithm.
        (default: 20 iterations).

    max_dist: positive float
        The algorithm uses an axis aligned bounding box (aabb) to avoid some distance computations.
        The algorithm skips a more precise distance computation if the distance between the aabb
        of the primitive objects composing the link and 'obj' is less than 'max_dist' (in meters).
        (default: infinite).

    Returns
    -------
    dist_struct : 'DistStructRobotObj' object
        Distance struct for each one of the m objects that compose the robot's
        collision model. Contains a list of m 'DistStructLinkObj' objects.
    """

        return _compute_dist(self, obj, q, htm, old_dist_struct, tol, no_iter_max, max_dist)

    def compute_dist_auto(self, q=None, old_dist_struct=None, tol=0.0005,
                     no_iter_max=20, max_dist = np.inf):
        """
    Compute the  distance structure from each one of the robot's links to itself
    (auto collision), given a joint and base configuration.

    This function consider only non-sequential links, since the collision between
    sequential links in the kinematic chain can be verified by checking joint limits.
    This saves times. This verification should be done elsewhere (by checking if the
    configuration is inside the joint limits).

    This function can be faster if some distance computations are avoided.
    See the description of the parameter 'max_dist'.

    Use an iterative algorithm, based on projections
    (Von Neumann's cyclic projection algorithm).

    Parameters
    ----------
    q : nd numpy vector or array
        The manipulator's joint configuration.
        (default: the current  joint configuration (robot.q) for the manipulator).

    old_dist_struct : 'DistStructRobotAuto' object
        'DistStructRobotAuto' obtained previously for the same robot.
        Can be used to enhance the algorithm speed using the previous closest
        point as an initial guess.
        (default: None).

    tol : positive float
        Tolerance for convergence in the iterative algorithm, in meters.
        (default: 0.0005 m).

    no_iter_max : positive int
        The maximum number of iterations for the algorithm.
        (default: 20 iterations).

    max_dist: positive float
        The algorithm uses an axis aligned bounding box (aabb) to avoid some distance computations.
        The algorithm skips a more precise distance computation if the distance between the aabb
        of the primitive objects composing the link and 'obj' is less than 'max_dist' (in meters).
        (default: infinite).

    Returns
    -------
    dist_struct : 'DistStructRobotAuto' object
        Distance struct for each one of the m objects that compose the robot's
        collision model. Contains a list of m 'DistStructLinkLink' objects.
    """

        return _compute_dist_auto(self, q, old_dist_struct, tol, no_iter_max, max_dist)

    def check_free_configuration(self, q=None, htm=None, obstacles=[],
                              check_joint=True, check_auto=True,
                              tol=0.0005, dist_tol=0.005, no_iter_max=20):
        """
    Check if the joint configuration q is in the free configuration space, considering
    joint limits, collision with obstacles and auto collision. It also outputs a message about a
    possible violation.

    For efficiency purposes, the program halts in the first violation it finds (if there is any).
    So, the message, if any, is only about one of the possible violations. There can be more.

    Parameters
    ----------
    q : nd numpy vector or array
        The manipulator's joint configuration.
        (default: the current joint configuration (robot.q) for the manipulator).

    obstacles : list of simple objects
        A list of obstacles as simple objects (see Utils.IS_SIMPLE)
        (default: empty list).

    htm : 4x4 numpy array or 4x4 nested list
        The robot base's configuration.
        (default: the same as the current htm).


    check_joint: boolean
        If joint limits should be considered or not.
        (default: True).

    check_auto: boolean
        If auto collision should be considered or not.
        (default: True).

    tol : positive float
        Tolerance for convergence in the iterative algorithm, in meters.
        (default: 0.0005 m).

    dist_tol : positive float
        The tolerance to consider that two links are colliding.
        (default: 0.005 m).

    no_iter_max : positive int
        The maximum number of iterations for the distance computing algorithm.
        (default: 20 iterations).

    Returns
    -------
    is_free : boolean
        If the configuration is free or not.

    message: string
        A message about what is colliding (is otherwise just 'Ok!').

    info: list
        Contains a list with information of the violation of free space (if there is any, otherwise
        it is empty). The first element is the violation type: either 0 (lower joint limit violated), 1 (upper joint
        limit violated) 2 (collision with obstacles) and 3 (collision between links). The last elements depends on
        the violation type.

        If lower joint limit or upper joint limit, contains which joint was violated.

        If collision with obstacles, contains the list [i,isub,j], containing the index of the link i
        i, the index of the collision object in the link isub and the obstacle index in the list, j.

        If collision, contains the list [i,isub,j,jsub], containing the index of the first link i,
        the index of the collision object in the first link isub, the index of the second link j and
        the index of the collision object in the second link jsub.


    """

        return _check_free_configuration(self, q, htm, obstacles, check_joint, check_auto, tol, dist_tol, no_iter_max)


from utils import *

from graphics.meshmaterial import *
from graphics.model3d import *

from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *

from .links import *


def _create_staubli_tx60(htm, name, color, opacity):
    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

    if not Utils.is_a_color(color):
        raise Exception("The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception("The parameter 'opacity' should be a float between 0 and 1.")

    c = (np.pi / 180)
    link_info = [[0, 0, 0, 0, 0, 0], #"theta" rotation in z
                 [0, 0, 0.02, 0.31, 0, 0.07],  # "d" translation in z
                 [3.14 / 2, 0, 3.14 / 2, 3.14 / 2, -3.14 / 2, 0],  # "alfa" rotation in x
                 [0, 0.29, 0, 0, 0, 0],  # "a" translation in x 0.25
                 [0, 0, 0, 0, 0, 0]] #joint type

    n = 6
    scale = 1

    # Collision model
    col_model = [[], [], [], [], [], []]

    # col_model[0].append(Box(htm=Utils.trn([-0.37, 0, 0]),
    #                             name=name + "_C0_0", width=0.25, depth=0.16, height=0.52, color="orange", opacity=0.3))


    #col_model[5].append(Ball(htm=np.identity(4),
    #                             name=name + "_C5_0", radius=0.0362, color="red", opacity=0.3))

    col_model[0].append(Cylinder(htm=Utils.trn([0, -0.15, 0]) @ Utils.rotx(-3.14 / 2),
                                 name=name + "_C0_0", radius=0.1, height=0.45, color="red", opacity=0.3))
    col_model[1].append(Box(htm=Utils.trn([-0.15, 0, -0.125]) @ Utils.rotz(3.14 / 2) @ Utils.rotx(-3.14 / 2 - c * 7.3),
                            name=name + "_C1_0", width=0.18, depth=0.1, height=0.45, color="blue", opacity=0.3))
    col_model[2].append(Box(htm=Utils.trn([0, -0.02, 0]),
                            name=name + "_C2_0", width=0.17, depth=0.2, height=0.25, color="green", opacity=0.3))
    col_model[3].append(Box(htm=Utils.trn([0, -0.08, 0]),
                            name=name + "_C3_0", width=0.15, depth=0.27, height=0.13, color="orange", opacity=0.3))

    # Create 3d objects

    base_3d_obj = [Model3D(
        'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/base_link.stl',
        scale,
        np.identity(4),
        MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5],
                     color=color, opacity=opacity))]

    link_3d_obj = []

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/link_1.stl',
            scale,
            Utils.trn([0, -0.005, 0]) @ Utils.rotx(-3.14 / 2),
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity))
         ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/link_2.stl',
            scale,
            Utils.trn([-0.31, 0, 0.02]) @ Utils.rotz(-3.14 / 2) @ Utils.rotx(-3.14 / 2 + c * 7.3),
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity))
         ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/link_3.stl',
            scale,
            Utils.rotz(-3.14) @ Utils.trn([0, 0.045 - 0.05, 0]),
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity))
         ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/link_4.stl',
            scale,
            Utils.trn([0, -0.31, 0]) @ Utils.roty(3.14) @ Utils.rotx(-3.14 / 2),
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                         opacity=opacity))
         ]
    )

    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/link_5.stl',
            scale,
            np.identity(4),
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="#C0C0C0",
                         opacity=opacity))
         ]
    )
    link_3d_obj.append(
        [Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/StaubliTX60/link_6.stl',
            scale,
            np.identity(4),
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="black",
                         opacity=opacity))
         ]
    )

    # Inertial parameter of the robots

    list_com_coordinates = [[0, -0.15, 0],
                            [-0.15, 0, -0.125],
                            [0, -0.02, 0],
                            [0, -0.08, 0],
                            [0.0, 0.0, 0.03],
                            [0.0, 0.0, 0.0]]

    list_inertia_mat = []

    list_inertia_mat.append(np.array([[0.8386, 0.0000, 0.0000],
                                      [0.0000, 0.1001, 0.0002],
                                      [0.0000, 0.0002, 0.8386]]) + np.diag([3.0e-4, 3.0e-4, 4.5e-4]))

    list_inertia_mat.append(np.array( [[ 0.2198, 0.0000, -0.2150],
                                       [ 0.0000, 0.6621, -0.0000],
                                       [-0.2150,-0.0000,  0.4614]]) + np.diag([3.0e-4, 3.0e-4, 4.5e-4]))

    list_inertia_mat.append(np.array( [[ 0.0965, -0.0000,  0.0000],
                                       [-0.0000,  0.0691,  0.0000],
                                       [ 0.0000,  0.0000,  0.1077]]) + np.diag([3.0e-4, 3.0e-4, 4.5e-4]))


    list_inertia_mat.append(np.array( [[ 0.1070, -0.0001, -0.0000],
                                       [-0.0001,  0.0245,  0.0000],
                                       [-0.0000,  0.0000,  0.1035]]) + np.diag([3.0e-4, 3.0e-4, 4.5e-4]))

    list_inertia_mat.append(np.array([[0., 0., 0.],
                                      [0., 0., 0.],
                                      [0., 0., 0.]]) + np.diag([3.0e-4, 3.0e-4, 4.5e-4]))

    list_inertia_mat.append(np.array([[0., 0., 0.],
                                      [0., 0., 0.],
                                      [0., 0., 0.]]) + np.diag([3.0e-4, 3.0e-4, 4.5e-4]))

    list_mass = [20.026, 11.474, 12.040, 7.458, 0.300, 0.300]

    # Create links

    links = []
    for i in range(n):
        links.append(Link(i, link_info[0][i], link_info[1][i], link_info[2][i], link_info[3][i], link_info[4][i],
                          link_3d_obj[i], list_mass[i], list_com_coordinates[i], list_inertia_mat[i]))

        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration

    htm_base_0 = Utils.trn([0, 0, 0.375])

    q0 = [0, 3.14/2, 0, 0, 0, 0]

    #Create joint limits
    joint_limits = (np.pi/180)*np.matrix([[-180,180],[-127.5+90, 127.5+90],[-142.5+90, 142.5+90],[-270,270],[-121,132.5],[-270,270]])

    return base_3d_obj, links, htm_base_0, np.identity(4), q0, joint_limits



from utils import *

from graphics.meshmaterial import *
from graphics.model3d import *

from simobjects.ball import *
from simobjects.box import *
from simobjects.cylinder import *

from .links import *


def _create_abb_crb(htm, name, color, opacity):
    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    if not (Utils.is_a_name(name)):
        raise Exception(
            "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

    if not Utils.is_a_color(color):
        raise Exception("The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception("The parameter 'opacity' should be a float between 0 and 1.")

    link_info = [[       0,     0,      3.14,         0, 0, 0],  # "theta" rotation in z
                 [   0.265,     0,         0,     -0.47, 0, 0.101],  # "d" translation in z
                 [3.14 / 2,     0, -3.14 / 2, -3.14 / 2, -3.14 / 2, 0],  # "alfa" rotation in x
                 [       0, 0.444,      0.11,         0, 0.08, 0],  # "a" translation in x
                 [       0,      0,        0,         0, 0, 0]]  # joint type

    scale = 1
    n = 6

    # Collision model
    col_model = [[], [], [], [], [], []]

    # Create 3d objects

    base_3d_obj = [
        Model3D(
            'https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/ABBCRB15000/base_link.stl',
            scale,
            np.identity(4),
            MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="#606060",
                         opacity=opacity))]

    link_3d_obj = []

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/ABBCRB15000/link_1.stl',
                 scale,
                 Utils.rotx(-3.14 / 2),
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="#606060",
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/ABBCRB15000/link_2.stl',
                 scale,
                 Utils.trn([-0.444, 0, 0]) @ Utils.rotx(-3.14 / 2) @ Utils.roty(3.14 / 2),
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="#606060",
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/ABBCRB15000/link_3.stl',
                 scale,
                 Utils.trn([-0.11, 0, 0]) @ Utils.roty(3.14 / 2),
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="#606060",
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/ABBCRB15000/link_4.stl',
                 scale,
                 Utils.trn([0, -0.47, 0]) @ Utils.roty(-3.14 / 2) @ Utils.rotz(3.14 / 2),
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/ABBCRB15000/link_5.stl',
                 scale,
                 Utils.trn([-0.08, 0, 0]) @ Utils.rotz(3.14 / 2) @ Utils.roty(-3.14 / 2) @ Utils.rotx(3.14 / 2),
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color=color,
                              opacity=opacity))]
    )

    link_3d_obj.append(
        [Model3D('https://cdn.jsdelivr.net/gh/viniciusmgn/uaibot_content@master/contents/ABBCRB15000/link_6.stl',
                 scale,
                 Utils.trn([0, 0, 0]) @ Utils.roty(-3.14 / 2),
                 MeshMaterial(metalness=0.7, clearcoat=1, roughness=0.5, normal_scale=[0.5, 0.5], color="silver",
                              opacity=opacity))]
    )

    # Inertial parameter of the robots

    # Create links

    links = []
    for i in range(n):
        links.append(Link(i, link_info[0][i], link_info[1][i], link_info[2][i], link_info[3][i], link_info[4][i],
                          link_3d_obj[i]))

        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration
    q0 = [0.000, 0.000, 3.14, 0.000, 0.000, 0.000]

    #Create joint limits
    joint_limits = (np.pi/180)*np.matrix([[-180,180],[-180,180],[-225-200+10,85-200-20],[-180,180],[-180,180],[-180,180]])

    return base_3d_obj, links, np.identity(4), np.identity(4), q0, joint_limits



from utils import *
import numpy as np
from graphics.meshmaterial import *


class SmoothBox:
    """
  A smooth box object. Smooth boxes satisfy the parametric equation (in their own aligned axis)

  (x/width)^4+(y/depth)^4+(z/height)^4 = 1

  Parameters
  ----------
  htm : 4x4 numpy array or 4x4 nested list
      The object's configuration
      (default: the same as the current HTM).

  name : string
      The object's name.
      (default: 'genBox').

  width : positive float
      The object's width, in meters.
      (default: 1).

  depth : positive float
      The object's depth, in meters.
      (default: 1).

  height : positive float
      The object's height, in meters.
      (default: 1).

  mass : positive float
      The object's mass, in kg.
      (default: 1).

  color : string
      The object's color, a HTML - compatible string.
      (default: "red").

  opacity : float between 0 and 1
      The opacity. 1 = fully opaque, and 0 = transparent.

  mesh_material: 'MeshMaterial' object
      The object mesh material. If set to 'None', the default is used.
      (default: None).
  """

    #######################################
    # Attributes
    #######################################

    @property
    def width(self):
        """The box width, in meters."""
        return self._width

    @property
    def height(self):
        """The box height, in meters."""
        return self._height

    @property
    def depth(self):
        """The box depth, in meters."""
        return self._depth

    @property
    def name(self):
        """The object name."""
        return self._name

    @property
    def htm(self):
        """Object pose. A 4x4 homogeneous transformation matrix written is scenario coordinates."""
        return np.matrix(self._htm)

    @property
    def mass(self):
        """Mass of the object, in kg."""
        return self._mass

    @property
    def color(self):
        """Color of the object"""
        return self.mesh_material.color

    @property
    def mesh_material(self):
        """Mesh properties of the object"""
        return self._mesh_material

    @property
    def volume(self):
        """The volume of the object, in m³."""
        return self._volume

    #######################################
    # Constructor
    #######################################

    def __init__(self, htm=np.identity(4), name="", width=1, height=1, depth=1, mass=1, color="red", opacity=1, \
                 mesh_material=None):

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_number(mass) or mass < 0:
            raise Exception("The parameter 'mass' should be a positive float.")

        if not Utils.is_a_number(width) or width < 0:
            raise Exception("The parameter 'width' should be a positive float.")

        if not Utils.is_a_number(height) or height < 0:
            raise Exception("The parameter 'height' should be a positive float.")

        if not Utils.is_a_number(depth) or depth < 0:
            raise Exception("The parameter 'depth' should be a positive float.")

        if name=="":
            name="var_smoothbox_id_"+str(id(self))

        if not (Utils.is_a_name(name)):
            raise Exception(
                "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

        if not Utils.is_a_color(color):
            raise Exception("The parameter 'color' should be a HTML-compatible color.")

        if not ((mesh_material is None) or (Utils.get_uaibot_type(mesh_material) == "uaibot.MeshMaterial")):
            raise Exception(
                "The parameter 'mesh_material' should be either 'None' or a 'uaibot.MeshMaterial' object.")

        if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
            raise Exception("The parameter 'opacity' should be a float between 0 and 1")
            # end error handling

        self._width = width
        self._height = height
        self._depth = depth
        self._htm = np.matrix(htm)
        self._name = name
        self._mass = 1
        self._frames = []
        self._volume = self.width * self.depth * self.height
        self._max_time = 0

        if mesh_material is None:
            self._mesh_material = MeshMaterial(color=color, opacity=opacity)
        else:
            self._mesh_material = mesh_material

        # Set initial total configuration
        self.set_ani_frame(self._htm)

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "SmoothBox with name '" + self.name + "': \n\n"
        string += " Width (m): " + str(self.width) + "\n"
        string += " Depth (m): " + str(self.depth) + "\n"
        string += " Height (m): " + str(self.height) + "\n"
        string += " Color: " + str(self.color) + "\n"
        string += " Mass (kg): " + str(self.mass) + "\n"
        string += " HTM: \n" + str(self.htm) + "\n"

        return string

    #######################################
    # Methods
    #######################################

    def add_ani_frame(self, time, htm=None):
        """
    Add a single configuration to the object's animation queue.

    Parameters
    ----------
    time: positive float
        The timestamp of the animation frame, in seconds.
    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """
        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_number(time) or time < 0:
            raise Exception("The parameter 'time' should be a positive float.")
        # end error handling

        f = [time, np.around(htm[0,0],4), np.around(htm[0,2],4), np.around(-htm[0,1],4), np.around(htm[0,3],4),
             np.around(htm[1,0],4), np.around(htm[1,2],4), np.around(-htm[1,1],4), np.around(htm[1,3],4),
             np.around(htm[2,0],4), np.around(htm[2,2],4), np.around(-htm[2,1],4), np.around(htm[2,3],4),
             0, 0, 0, 1]

        self._htm = htm
        self._frames.append(f)
        self._max_time = max(self._max_time, time)

    def set_ani_frame(self, htm=None):
        """
    Reset object's animation queue and add a single configuration to the
    object's animation queue.

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """

        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        # end error handling

        self._frames = []
        self.add_ani_frame(0, htm)
        self._max_time = 0

    def gen_code(self):
        """Generate code for injection."""

        string = "\n"
        string += "//BEGIN DECLARATION OF THE SMOOTH BOX '" + self.name + "'\n\n"
        string += self.mesh_material.gen_code(self.name) + "\n"
        string += "const var_" + self.name + " = new SmoothBox(" + str(self.width) + "," + str(
            self.height) + "," + str(self.depth) + "," + str(self._frames) + ", material_" + self.name + ");\n"
        string += "sceneElements.push(var_" + self.name + ");\n"
        string += "//USER INPUT GOES HERE"
        return string

    # Compute inertia matrix with respect to the inertia frame
    def inertia_matrix(self, htm=None):
        """
    The 3D inertia matrix of the object, written in the world frame.
    Assume that the transformation between the word frame and the object frame is 'htm'.

    Parameters
    ----------
    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration for which the inertia matrix will be computed
        (default: the same as the current HTM).

    Returns
    -------
     inertia_matrix : 3x3 numpy array
        The 3D inertia matrix.
    """

        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The optional parameter 'htm' should be a 4x4 homogeneous transformation matrix.")
        # end error handling

        Ixx = 0.05486 * self.mass * (self.height * self.height + self.depth * self.depth)
        Iyy = 0.05486 * self.mass * (self.width * self.width + self.depth * self.depth)
        Izz = 0.05486 * self.mass * (self.height * self.height + self.width * self.width)
        Q = htm[0:3, 0:3]
        S = Utils.S(htm[0:3, 3])

        return Q * np.diag([Ixx, Iyy, Izz]) * Q.T - self.mass * S * S

    def copy(self):
        """Return a deep copy of the object, without copying the animation frames."""
        return SmoothBox(self.htm, self.name + "_copy", self.width, self.height, self.depth, self.mass, self.color)

    def aabb(self):
        """
    Compute the width, depth and height of an axis aligned bounding box (aabb) that
    covers the object. It also considers the current orientation.

    Returns
    -------
     width : positive float
        The width of the box, in meters.

     depth : positive float
        The depth of the box, in meters.

     height : positive float
        The depth of the box, in meters.
    """

        p1 = self.width * self.htm[:,0] + self.depth * self.htm[:,1] + self.height * self.htm[:,2]
        p2 = -self.width * self.htm[:, 0] + self.depth * self.htm[:, 1] + self.height * self.htm[:, 2]
        p3 = self.width * self.htm[:, 0] - self.depth * self.htm[:, 1] + self.height * self.htm[:, 2]
        p4 = self.width * self.htm[:, 0] + self.depth * self.htm[:, 1] - self.height * self.htm[:, 2]

        w = np.max([abs(p1[0, 0]), abs(p2[0, 0]), abs(p3[0, 0]), abs(p4[0, 0])])
        d = np.max([abs(p1[1, 0]), abs(p2[1, 0]), abs(p3[1, 0]), abs(p4[1, 0])])
        h = np.max([abs(p1[2, 0]), abs(p2[2, 0]), abs(p3[2, 0]), abs(p4[2, 0])])

        return w, d, h

    # Compute the projection of a point into an object
    def projection(self, point, htm=None):
        """
    The projection of a point in the object, that is, the
    closest point in the object to a point 'point'.

    Parameters
    ----------
    point : 3D vector
        The point for which the projection will be computed.

    htm : 4x4 numpy array or 4x4 nested list
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
     proj_point : 3D vector
        The projection of the point 'point' in the object.

     d : positive float
        The distance between the object and 'point'.
    """

        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The optional parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_vector(point, 3):
            raise Exception("The parameter 'point' should be a 3D vector.")

        # end error handling
        tpoint = htm[0:3, 0:3].T * (point - htm[0:3, 3])

        A = self.width/2
        B = self.depth/2
        C = self.height/2

        d4 = (tpoint[0, 0]/A)**4 + (tpoint[1, 0]/B)**4 + (tpoint[2, 0]/C)**4

        if d4 <= 1:
            return point, 0
        else:

            fun = lambda lam: (Utils.cubicsolve(tpoint[0, 0], lam / (A ** 4)) / A) ** 4 + \
                              (Utils.cubicsolve(tpoint[1, 0], lam / (B ** 4)) / B) ** 4 + \
                              (Utils.cubicsolve(tpoint[2, 0], lam / (C ** 4)) / C) ** 4 - 1

            maxR = max(A, B, C)
            minR = min(A, B, C)
            dp = np.linalg.norm(tpoint)

            lam0 = 0.00001
            lamf = maxR*(dp-minR)/0.75

            lam_star = Utils.bissection(fun,lam0,lamf,0.0001)

            x = Utils.cubicsolve(tpoint[0, 0], lam_star / (A ** 4))
            y = Utils.cubicsolve(tpoint[1, 0], lam_star / (B ** 4))
            z = Utils.cubicsolve(tpoint[2, 0], lam_star / (C ** 4))

            d = sqrt( (x-tpoint[0,0])**2 + (y-tpoint[1,0])**2 + (z-tpoint[2,0])**2 )

            return htm[0:3, 0:3] * np.matrix([[x], [y], [z]]) + htm[0:3, 3], d
